########################################################################################################################
# Module: tests/test_core.py
# Description: Tests for core and Sampler
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import unittest

import jax.numpy as jnp
import mocat.src.sample
import numpy.testing as npt

from mocat.src import core
from mocat.src import sample


class Testcdict(unittest.TestCase):
    cdict = core.cdict(test_arr=jnp.ones((10, 3)),
                       test_float=3.)

    def test_init(self):
        npt.assert_(hasattr(self.cdict, 'test_arr'))
        npt.assert_array_equal(self.cdict.test_arr, jnp.ones((10, 3)))

        npt.assert_(hasattr(self.cdict, 'test_float'))
        npt.assert_equal(self.cdict.test_float, 3.)

    def test_copy(self):
        cdict2 = self.cdict.copy()
        npt.assert_(isinstance(cdict2, core.cdict))

        npt.assert_(isinstance(cdict2.test_arr, jnp.DeviceArray))
        npt.assert_array_equal(cdict2.test_arr, jnp.ones((10, 3)))

        npt.assert_(isinstance(cdict2.test_float, float))
        npt.assert_equal(cdict2.test_float, 3.)

        cdict2.test_arr = jnp.zeros(5)
        npt.assert_array_equal(self.cdict.test_arr, jnp.ones((10, 3)))

        cdict2.test_float = 9.
        npt.assert_equal(self.cdict.test_float, 3.)

    def test_getitem(self):
        cdict_0get = self.cdict[0]
        npt.assert_(isinstance(cdict_0get, core.cdict))

        npt.assert_(isinstance(cdict_0get.test_arr, jnp.DeviceArray))
        npt.assert_array_equal(cdict_0get.test_arr, jnp.ones(3))

        npt.assert_(isinstance(cdict_0get.test_float, float))
        npt.assert_equal(cdict_0get.test_float, 3.)

    def test_additem(self):
        cdict_other = core.cdict(test_arr=jnp.ones((2, 3)),
                                 test_float=7.,
                                 time=25.)

        self.cdict.time = 10.

        cdict_add = self.cdict + cdict_other

        npt.assert_(isinstance(cdict_add, core.cdict))

        npt.assert_(isinstance(cdict_add.test_arr, jnp.DeviceArray))
        npt.assert_array_equal(cdict_add.test_arr, jnp.ones((12, 3)))
        npt.assert_array_equal(cdict_add.time, 35.)

        npt.assert_(isinstance(cdict_add.test_float, float))
        npt.assert_equal(cdict_add.test_float, 3.)

        npt.assert_array_equal(self.cdict.test_arr, jnp.ones((10, 3)))
        npt.assert_equal(self.cdict.test_float, 3.)
        npt.assert_equal(self.cdict.time, 10.)
        del self.cdict.time


class TestSampler(unittest.TestCase):
    sampler = sample.Sampler(name='test', other=jnp.zeros(2))

    def test_init(self):
        npt.assert_equal(self.sampler.name, 'test')
        npt.assert_(hasattr(self.sampler, 'parameters'))
        npt.assert_array_equal(self.sampler.parameters.other, jnp.zeros(2))

    def test_copy(self):
        sampler2 = self.sampler.deepcopy()

        npt.assert_(isinstance(sampler2, sample.Sampler))

        sampler2.name = 'other'
        npt.assert_equal(self.sampler.name, 'test')

        sampler2.parameters.other = 10.
        npt.assert_array_equal(self.sampler.parameters.other, jnp.zeros(2))


if __name__ == '__main__':
    unittest.main()



########################################################################################################################
# Module: ssm/backward.py
# Description: Backward simulation for state-space models.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple
from time import time
from functools import partial

from jax import numpy as jnp, random, vmap, jit
from jax.lax import while_loop, scan, cond, map

from mocat.src.core import cdict
from mocat.src.ssm.ssm import StateSpaceModel
from mocat.src.ssm.filtering import ParticleFilter, run_particle_filter_for_marginals


def full_resample_single(ssm_scenario: StateSpaceModel,
                         x0_all: jnp.ndarray,
                         t: float,
                         x1_single: jnp.ndarray,
                         tplus1: float,
                         x0_log_weight: jnp.ndarray,
                         random_key: jnp.ndarray) -> jnp.ndarray:
    log_weight = x0_log_weight - vmap(ssm_scenario.transition_potential, (0, None, None, None))(x0_all, t,
                                                                                                x1_single, tplus1)
    return x0_all[random.categorical(random_key, log_weight)]


def full_resampling(ssm_scenario: StateSpaceModel,
                    x0_all: jnp.ndarray,
                    t: float,
                    x1_all: jnp.ndarray,
                    tplus1: float,
                    x0_log_weight: jnp.ndarray,
                    random_key: jnp.ndarray) -> jnp.ndarray:
    return vmap(full_resample_single, (None, None, None, 0, None, None, 0)) \
        (ssm_scenario, x0_all, t, x1_all, tplus1, x0_log_weight, random.split(random_key, len(x1_all)))


def full_resample_single_cond(not_yet_accepted: bool,
                              x0_false: jnp.ndarray,
                              ssm_scenario: StateSpaceModel,
                              x0_all: jnp.ndarray,
                              t: float,
                              x1_single: jnp.ndarray,
                              tplus1: float,
                              x0_log_weight: jnp.ndarray,
                              random_key: jnp.ndarray) -> jnp.ndarray:
    return cond(not_yet_accepted,
                lambda _: full_resample_single(ssm_scenario, x0_all, t, x1_single, tplus1, x0_log_weight, random_key),
                lambda _: x0_false,
                None)


def rejection_proposal_single(ssm_scenario: StateSpaceModel,
                              x0_all: jnp.ndarray,
                              t: float,
                              x1_single: jnp.ndarray,
                              tplus1: float,
                              x0_log_weight: jnp.ndarray,
                              bound: float,
                              random_key: jnp.ndarray) \
        -> Tuple[jnp.ndarray, float, bool, jnp.ndarray]:
    random_key, choice_key, uniform_key = random.split(random_key, 3)
    x0_single = x0_all[random.categorical(choice_key, x0_log_weight)]
    conditional_dens = jnp.exp(-ssm_scenario.transition_potential(x0_single, t, x1_single, tplus1))
    return x0_single, conditional_dens, random.uniform(uniform_key) > conditional_dens / bound, random_key


def rejection_proposal_single_cond(not_yet_accepted: bool,
                                   x0_false: jnp.ndarray,
                                   ssm_scenario: StateSpaceModel,
                                   x0_all: jnp.ndarray,
                                   t: float,
                                   x1_single: jnp.ndarray,
                                   tplus1: float,
                                   x0_log_weight: jnp.ndarray,
                                   bound: float,
                                   random_key: jnp.ndarray) -> Tuple[jnp.ndarray, float, bool, jnp.ndarray]:
    return cond(not_yet_accepted,
                lambda _: rejection_proposal_single(ssm_scenario,
                                                    x0_all, t,
                                                    x1_single, tplus1,
                                                    x0_log_weight, bound, random_key),
                lambda _: (x0_false, 0., False, random_key),
                None)


def rejection_proposal_all(ssm_scenario: StateSpaceModel,
                           x0_all: jnp.ndarray,
                           t: float,
                           x1_all: jnp.ndarray,
                           tplus1: float,
                           x0_log_weight: jnp.ndarray,
                           bound_inflation: float,
                           not_yet_accepted_arr: jnp.ndarray,
                           x0_all_sampled: jnp.ndarray,
                           bound: float,
                           random_keys: jnp.ndarray,
                           rejection_iter: int,
                           num_transition_evals: int) -> Tuple[jnp.ndarray, jnp.ndarray, float, jnp.ndarray, int, int]:
    n = len(x1_all)
    mapped_tup = map(lambda i: rejection_proposal_single_cond(not_yet_accepted_arr[i],
                                                              x0_all_sampled[i],
                                                              ssm_scenario,
                                                              x0_all,
                                                              t,
                                                              x1_all[i],
                                                              tplus1,
                                                              x0_log_weight,
                                                              bound,
                                                              random_keys[i]), jnp.arange(n))
    x0_all_sampled, dens_evals, not_yet_accepted_arr_new, random_keys = mapped_tup

    # Check if we need to start again
    max_dens = jnp.max(dens_evals)
    reset_bound = max_dens > bound
    bound = jnp.where(reset_bound, max_dens * bound_inflation, bound)
    not_yet_accepted_arr_new = jnp.where(reset_bound, jnp.ones(n, dtype='bool'), not_yet_accepted_arr_new)
    return not_yet_accepted_arr_new, x0_all_sampled, bound, random_keys, rejection_iter + 1, \
           num_transition_evals + not_yet_accepted_arr.sum()


def rejection_resampling(ssm_scenario: StateSpaceModel,
                         x0_all: jnp.ndarray,
                         t: float,
                         x1_all: jnp.ndarray,
                         tplus1: float,
                         x0_log_weight: jnp.ndarray,
                         random_key: jnp.ndarray,
                         maximum_rejections: int,
                         init_bound_param: float,
                         bound_inflation: float) -> Tuple[jnp.ndarray, int]:
    rejection_initial_keys = random.split(random_key, 3)
    n = len(x1_all)

    # Prerun to initiate bound
    x0_initial = x0_all[random.categorical(rejection_initial_keys[0], x0_log_weight, shape=(n,))]
    initial_cond_dens = jnp.exp(-vmap(ssm_scenario.transition_potential,
                                      (0, None, 0, None))(x0_initial, t, x1_all, tplus1))
    max_cond_dens = jnp.max(initial_cond_dens)
    initial_bound = jnp.where(max_cond_dens > init_bound_param, max_cond_dens * bound_inflation, init_bound_param)
    initial_not_yet_accepted_arr = random.uniform(rejection_initial_keys[1], (n,)) > initial_cond_dens / initial_bound

    out_tup = while_loop(lambda tup: jnp.logical_and(tup[0].sum() > 0, tup[-2] < maximum_rejections),
                         lambda tup: rejection_proposal_all(ssm_scenario, x0_all, t, x1_all, tplus1, x0_log_weight,
                                                            bound_inflation, *tup),
                         (initial_not_yet_accepted_arr,
                          x0_initial,
                          initial_bound,
                          random.split(rejection_initial_keys[2], n),
                          1,
                          n))
    not_yet_accepted_arr, final_particles, final_bound, random_keys, rej_attempted, num_transition_evals = out_tup

    final_particles = map(lambda i: full_resample_single_cond(not_yet_accepted_arr[i],
                                                              final_particles[i],
                                                              ssm_scenario,
                                                              x0_all,
                                                              t,
                                                              x1_all[i],
                                                              tplus1,
                                                              x0_log_weight,
                                                              random_keys[i]), jnp.arange(n))

    num_transition_evals = num_transition_evals + len(x0_all) * not_yet_accepted_arr.sum()

    return final_particles, num_transition_evals


@partial(jit, static_argnums=(0, 2))
def backward_simulation_rejection(ssm_scenario: StateSpaceModel,
                                  marginal_particles: cdict,
                                  n_samps: int,
                                  random_key: jnp.ndarray,
                                  maximum_rejections: int,
                                  init_bound_param: float,
                                  bound_inflation: float) -> cdict:
    marg_particles_vals = marginal_particles.value
    times = marginal_particles.t
    marginal_log_weight = marginal_particles.log_weight

    T, n_pf, d = marg_particles_vals.shape

    t_keys = random.split(random_key, T)
    final_particle_vals = marg_particles_vals[-1, random.categorical(t_keys[-1],
                                                                     marginal_log_weight[-1],
                                                                     shape=(n_samps,))]

    def back_sim_body(x_tplus1_all: jnp.ndarray, ind: int):
        x_t_all, num_transition_evals = rejection_resampling(ssm_scenario,
                                                             marg_particles_vals[ind], times[ind],
                                                             x_tplus1_all, times[ind + 1],
                                                             marginal_log_weight[ind], t_keys[ind],
                                                             maximum_rejections, init_bound_param, bound_inflation)
        return x_t_all, (x_t_all, num_transition_evals)

    _, back_sim_out = scan(back_sim_body,
                           final_particle_vals,
                           jnp.arange(T - 2, -1, -1), unroll=1)

    back_sim_particles, num_transition_evals = back_sim_out

    out_samps = marginal_particles.copy()
    out_samps.value = jnp.vstack([back_sim_particles[::-1], final_particle_vals[jnp.newaxis]])
    out_samps.num_transition_evals = jnp.append(0, num_transition_evals[::-1])
    del out_samps.log_weight
    return out_samps


@partial(jit, static_argnums=(0, 2))
def backward_simulation_full(ssm_scenario: StateSpaceModel,
                             marginal_particles: cdict,
                             n_samps: int,
                             random_key: jnp.ndarray) -> cdict:
    marg_particles_vals = marginal_particles.value
    times = marginal_particles.t
    marginal_log_weight = marginal_particles.log_weight

    T, n_pf, d = marg_particles_vals.shape

    t_keys = random.split(random_key, T)
    final_particle_vals = marg_particles_vals[-1, random.categorical(t_keys[-1],
                                                                     marginal_log_weight[-1],
                                                                     shape=(n_samps,))]

    def back_sim_body(x_tplus1_all: jnp.ndarray, ind: int):
        x_t_all = full_resampling(ssm_scenario, marg_particles_vals[ind], times[ind],
                                  x_tplus1_all, times[ind + 1], marginal_log_weight[ind], t_keys[ind])
        return x_t_all, x_t_all

    _, back_sim_particles = scan(back_sim_body,
                                 final_particle_vals,
                                 jnp.arange(T - 2, -1, -1))

    out_samps = marginal_particles.copy()
    out_samps.value = jnp.vstack([back_sim_particles[::-1], final_particle_vals[jnp.newaxis]])
    out_samps.num_transition_evals = jnp.append(0, jnp.ones(T - 1) * n_pf * n_samps)
    del out_samps.log_weight
    return out_samps


def backward_simulation(ssm_scenario: StateSpaceModel,
                        marginal_particles: cdict,
                        random_key: jnp.ndarray,
                        n_samps: int = None,
                        maximum_rejections: int = 0,
                        transition_dens_bound_parameter: float = 0.,
                        bound_inflation: float = 1.01) -> cdict:
    if n_samps is None:
        n_samps = marginal_particles.value.shape[1]

    if maximum_rejections > 0:
        bsi_samps = backward_simulation_rejection(ssm_scenario,
                                                  marginal_particles,
                                                  n_samps,
                                                  random_key,
                                                  maximum_rejections,
                                                  transition_dens_bound_parameter,
                                                  bound_inflation)
    else:
        bsi_samps = backward_simulation_full(ssm_scenario,
                                             marginal_particles,
                                             n_samps,
                                             random_key)

    return bsi_samps


def forward_filtering_backward_simulation(ssm_scenario: StateSpaceModel,
                                          particle_filter: ParticleFilter,
                                          y: jnp.ndarray,
                                          t: jnp.ndarray,
                                          n_samps: int,
                                          random_key: jnp.ndarray,
                                          n_pf: int = None,
                                          ess_threshold: float = 0.5,
                                          maximum_rejections: int = 0,
                                          transition_dens_bound_parameter: float = 0.,
                                          bound_inflation: float = 1.01) -> cdict:
    pf_key, bsi_key = random.split(random_key)

    if n_pf is None:
        n_pf = n_samps

    pf_start = time()
    pf_samps = run_particle_filter_for_marginals(ssm_scenario,
                                                 particle_filter,
                                                 y,
                                                 t,
                                                 pf_key,
                                                 n=n_pf,
                                                 ess_threshold=ess_threshold)
    pf_samps.value.block_until_ready()
    pf_end = time()

    bsi_samps = backward_simulation(ssm_scenario,
                                    pf_samps,
                                    bsi_key,
                                    n_samps,
                                    maximum_rejections,
                                    transition_dens_bound_parameter,
                                    bound_inflation)

    bsi_samps.value.block_until_ready()
    bsi_end = time()

    bsi_samps.time = bsi_end - pf_start
    bsi_samps.pf_time = pf_end - pf_start
    bsi_samps.bsi_time = bsi_end - pf_end

    return bsi_samps



########################################################################################################################
# Module: tests/test_transport.py
# Description: Tests for SVGD and SMC samplers.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################


import unittest
from typing import Tuple

from jax import numpy as jnp, random, vmap
import numpy.testing as npt

from mocat.src.core import cdict
from mocat.src.sample import run
from mocat.src.scenarios.twodim import toy_examples
from mocat.src.transport.smc import MetropolisedSMCSampler
from mocat.src.transport.svgd import SVGD
from mocat.src.kernels import median_bandwidth_update, mean_bandwidth_update
from mocat.src.mcmc.standard_mcmc import RandomWalk, Underdamped


class TestCorrelatedGaussian(unittest.TestCase):
    n: int
    scenario = toy_examples.Gaussian(covariance=jnp.array([[1., 0.9], [0.9, 2.]]))
    scenario.prior_sample = lambda rk: (random.normal(rk, shape=(2,)) * 7.)
    scenario.prior_potential = lambda x, rk: 0.5 * jnp.square(x / 7. ** 2).sum(-1)
    posterior_covariance = jnp.linalg.inv(jnp.linalg.inv(scenario.covariance) + 1/7**2 * jnp.eye(2))

    def _test_mean(self,
                   sample: cdict):
        if sample.value.ndim == 3:
            val = jnp.concatenate(sample.value)
        else:
            val = sample.value
        samp_mean = val.mean(axis=0)
        npt.assert_array_almost_equal(samp_mean, jnp.zeros(2), decimal=0)

    def _test_cov(self,
                  sample: cdict):
        if sample.value.ndim == 3:
            val = jnp.concatenate(sample.value)
        else:
            val = sample.value
        samp_cov = jnp.cov(val.T)
        npt.assert_array_almost_equal(samp_cov, self.posterior_covariance, decimal=1)

    def _test_log_norm_const(self,
                             sample):
        lik_prec = jnp.linalg.inv(self.scenario.covariance)
        tempered_covs = vmap(lambda t: jnp.linalg.inv(lik_prec * t + jnp.eye(2) / 7 ** 2))(sample.temperature)
        tempered_dets = vmap(jnp.linalg.det)(tempered_covs)

        npt.assert_array_almost_equal(sample.log_norm_constant, 0.5 * (jnp.log(tempered_dets) - 4 * jnp.log(7)), 0)

    def resample_final(self,
                       sample: cdict) -> cdict:
        unweighted_vals = sample.value[-1, random.categorical(random.PRNGKey(1),
                                                              logits=sample.log_weight[-1], shape=(self.n,))]
        unweighted_sample = cdict(value=unweighted_vals)
        return unweighted_sample


class TestSVGD(TestCorrelatedGaussian):
    n = int(1e2)
    n_iter = int(1e3)

    def test_fixed_kernel_params(self):
        sample = run(self.scenario, SVGD(max_iter=self.n_iter, stepsize=0.8),
                     n=self.n,
                     random_key=random.PRNGKey(0))
        self._test_mean(sample[-1])
        self._test_cov(sample[-1])

    def test_median_update(self):
        class SVGD_median(SVGD):
            def adapt(self,
                      ensemble_state: cdict,
                      ensemble_extra: cdict) -> Tuple[cdict, cdict]:
                ensemble_extra.parameters.kernel_params.bandwidth = median_bandwidth_update(ensemble_state.value)
                ensemble_state.kernel_params = ensemble_extra.parameters.kernel_params
                return ensemble_state, ensemble_extra

        sample = run(self.scenario, SVGD_median(max_iter=self.n_iter, stepsize=1.0),
                     n=self.n,
                     random_key=random.PRNGKey(0))
        self._test_mean(sample[-1])
        self._test_cov(sample[-1])

    def test_mean_update(self):
        class SVGD_mean(SVGD):
            def adapt(self,
                      ensemble_state: cdict,
                      ensemble_extra: cdict) -> Tuple[cdict, cdict]:
                ensemble_extra.parameters.kernel_params.bandwidth = mean_bandwidth_update(ensemble_state.value)
                ensemble_state.kernel_params = ensemble_extra.parameters.kernel_params
                return ensemble_state, ensemble_extra

        sample = run(self.scenario, SVGD_mean(max_iter=self.n_iter, stepsize=1.0),
                     n=self.n,
                     random_key=random.PRNGKey(0))
        self._test_mean(sample[-1])
        self._test_cov(sample[-1])

    def test_callable_stepsize(self):
        sample = run(self.scenario, SVGD(max_iter=self.n_iter, stepsize=lambda i: 10 * i ** -0.5),
                     n=self.n,
                     random_key=random.PRNGKey(0))
        self._test_mean(sample[-1])
        self._test_cov(sample[-1])

    def test_ensemble_minibatch(self):
        sample = run(self.scenario, SVGD(max_iter=self.n_iter, stepsize=1.0, ensemble_batchsize=100),
                     n=self.n,
                     random_key=random.PRNGKey(0))
        self._test_mean(sample[-1])
        self._test_cov(sample[-1])


class TestMetropolisedSMC(TestCorrelatedGaussian):
    preschedule = jnp.arange(0., 1.1, 0.1)
    n = int(1e4)

    def test_tempered_preschedule_RW(self):
        sample = run(self.scenario, MetropolisedSMCSampler(RandomWalk(stepsize=1.0)),
                     self.n,
                     random_key=random.PRNGKey(0),
                     temperature_schedule=self.preschedule)
        unweighted_sample = self.resample_final(sample)
        self._test_mean(unweighted_sample)
        self._test_cov(unweighted_sample)
        self._test_log_norm_const(sample)
        npt.assert_array_equal(sample.temperature, self.preschedule[1:])

    def test_tempered_preschedule_UD(self):
        sample = run(self.scenario, MetropolisedSMCSampler(Underdamped(stepsize=1.0)),
                     self.n,
                     random_key=random.PRNGKey(0),
                     temperature_schedule=self.preschedule)
        unweighted_sample = self.resample_final(sample)
        self._test_mean(unweighted_sample)
        self._test_cov(unweighted_sample)
        self._test_log_norm_const(sample)
        npt.assert_array_equal(sample.temperature, self.preschedule[1:])

    def test_tempered_adaptive_RW(self):
        sample = run(self.scenario, MetropolisedSMCSampler(RandomWalk(stepsize=1.0)),
                     self.n,
                     random_key=random.PRNGKey(0))
        unweighted_sample = self.resample_final(sample)
        self._test_mean(unweighted_sample)
        self._test_cov(unweighted_sample)
        self._test_log_norm_const(sample)

    def test_tempered_adaptive_UD(self):
        sample = run(self.scenario, MetropolisedSMCSampler(Underdamped(stepsize=1.0, leapfrog_steps=10)),
                     self.n,
                     random_key=random.PRNGKey(0))
        unweighted_sample = self.resample_final(sample)
        self._test_mean(unweighted_sample)
        self._test_cov(unweighted_sample)
        self._test_log_norm_const(sample)


if __name__ == '__main__':
    unittest.main()



########################################################################################################################
# Module: mcmc/metropolis.py
# Description: Accept/reject corrections for MCMC samplers.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple, Union, Type
from inspect import isclass

from jax import random, numpy as jnp
from jax.lax import cond

from mocat.src.core import cdict, Scenario
from mocat.src.mcmc.sampler import MCMCSampler, Correction


def mh_acceptance_probability(sampler: MCMCSampler,
                              scenario: Scenario,
                              reject_state: cdict, reject_extra: cdict,
                              proposed_state: cdict, proposed_extra: cdict) -> Union[float, jnp.ndarray]:
    pre_min_alpha = jnp.exp(- proposed_state.potential
                            + reject_state.potential
                            - sampler.proposal_potential(scenario,
                                                         proposed_state, proposed_extra,
                                                         reject_state, reject_extra)
                            + sampler.proposal_potential(scenario,
                                                         reject_state, reject_extra,
                                                         proposed_state, proposed_extra))

    return jnp.minimum(1., pre_min_alpha)


class Metropolis(Correction):

    def startup(self,
                scenario: Scenario,
                sampler: MCMCSampler,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        initial_state, initial_extra = super().startup(scenario, sampler, n,
                                                       initial_state, initial_extra, **kwargs)
        initial_state.alpha = 1.
        return initial_state, initial_extra

    def correct(self,
                scenario: Scenario,
                sampler: MCMCSampler,
                reject_state: cdict,
                reject_extra: cdict,
                proposed_state: cdict,
                proposed_extra: cdict) -> Tuple[cdict, cdict]:
        alpha = sampler.acceptance_probability(scenario,
                                               reject_state, reject_extra,
                                               proposed_state, proposed_extra)
        alpha = jnp.where(jnp.isnan(alpha), 0., alpha)

        random_key, subkey = random.split(proposed_extra.random_key)
        u = random.uniform(subkey)

        new_state, new_extra = cond(u < alpha,
                                    lambda _: (proposed_state, proposed_extra),
                                    lambda _: (reject_state, reject_extra),
                                    None)

        new_state.alpha = alpha
        new_extra.random_key = random_key
        return new_state, new_extra


class RMMetropolis(Correction):

    def __init__(self,
                 super_correction: Union[Correction, Type[Correction]] = Metropolis(),
                 adapt_cut_off: int = jnp.inf,
                 rm_stepsize_scale: float = 1.,
                 rm_stepsize_neg_exponent: float = 2 / 3,
                 log_update: bool = True):
        super().__init__()
        self.super_correction = super_correction() if isclass(super_correction) else super_correction
        self.super_correction.__init__()
        self.adapt_cut_off = adapt_cut_off
        self.rm_stepsize_scale = rm_stepsize_scale
        self.rm_stepsize_neg_exponent = rm_stepsize_neg_exponent
        self.log_update = log_update
        self._param_update = self.log_robbins_monro_update if self.log_update else self.robbins_monro_update
        self.tuning = None

    def startup(self,
                scenario: Scenario,
                sampler: MCMCSampler,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        initial_state, initial_extra = super().startup(scenario, sampler, n,
                                                       initial_state, initial_extra, **kwargs)
        # Set tuning parameter (i.e. stepsize) to 2.38^2/d if not initiated and adaptive
        if hasattr(sampler.parameters, sampler.tuning.parameter) \
                and getattr(sampler.parameters, sampler.tuning.parameter) is None:
            setattr(initial_extra.parameters, sampler.tuning.parameter, 2.38 ** 2 / scenario.dim)

        setattr(initial_state, sampler.tuning.parameter, getattr(initial_extra.parameters,
                                                                 sampler.tuning.parameter))

        initial_state, initial_extra = self.super_correction.startup(scenario, sampler, n,
                                                                     initial_state, initial_extra, **kwargs)

        sampler.tuning.monotonicity = 1 if sampler.tuning.monotonicity in (1, 'increasing') else -1

        self.tuning = sampler.tuning

        self._param_update = self.log_robbins_monro_update if self.log_update else self.robbins_monro_update
        return initial_state, initial_extra

    def correct(self,
                scenario: Scenario,
                sampler: MCMCSampler,
                reject_state: cdict,
                reject_extra: cdict,
                proposed_state: cdict,
                proposed_extra: cdict) -> Tuple[cdict, cdict]:
        corrected_state, corrected_extra = self.super_correction.correct(scenario, sampler,
                                                                         reject_state, reject_extra,
                                                                         proposed_state, proposed_extra)

        adapted_state, adapted_extra = cond(corrected_extra.iter <= self.adapt_cut_off,
                                            lambda carry: self.adapt(*carry),
                                            lambda carry: carry,
                                            (corrected_state, corrected_extra))

        setattr(adapted_state, sampler.tuning.parameter, getattr(adapted_extra.parameters,
                                                                 sampler.tuning.parameter))
        return adapted_state, adapted_extra

    def adapt(self,
              state: cdict,
              extra: cdict) -> Tuple[cdict, cdict]:
        param = extra.parameters.__dict__[self.tuning.parameter]
        param = self._param_update(param, getattr(state, self.tuning.metric),
                                   self.tuning.target, self.tuning.monotonicity,
                                   self.rm_stepsize_scale
                                   * extra.iter ** -self.rm_stepsize_neg_exponent)
        extra.parameters.__dict__[self.tuning.parameter] = param
        return state, extra

    @staticmethod
    def robbins_monro_update(param, metric_value, target, monotonicity, rm_stepsize):
        return param - rm_stepsize * (metric_value - target) * monotonicity

    @staticmethod
    def log_robbins_monro_update(param, metric_value, target, monotonicity, rm_stepsize):
        log_param = jnp.log(param)
        new_log_param = RMMetropolis.robbins_monro_update(log_param,
                                                          metric_value,
                                                          target,
                                                          monotonicity,
                                                          rm_stepsize)
        return jnp.exp(new_log_param)



########################################################################################################################
# Module: ssm/scenarios/stochastic_volatility.py
# Description: Multivariate stochastic volatility model.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union, Any

from jax import numpy as jnp, random
from mocat.src.utils import gaussian_potential, reset_covariance
from mocat.src.ssm.ssm import StateSpaceModel


class StochasticVolatility(StateSpaceModel):
    name = 'Stochastic Volatility'

    initial_covariance_sqrt = None
    initial_precision_sqrt = None
    transition_covariance_sqrt = None
    transition_precision_sqrt = None
    likelihood_covariance_diag_sqrt = None
    likelihood_precision_diag_sqrt = None

    def __init__(self,
                 dim: int,
                 initial_mean: jnp.ndarray = None,
                 initial_covariance: jnp.ndarray = None,
                 transition_matrix: jnp.ndarray = None,
                 transition_covariance: jnp.ndarray = None,
                 likelihood_covariance_diag: jnp.ndarray = None,
                 name: str = None):
        self.dim = dim
        self.dim_obs = dim
        if initial_mean is None:
            initial_mean = jnp.zeros(self.dim)
        self.initial_mean = initial_mean

        if initial_covariance is None:
            initial_covariance = jnp.eye(self.dim)
        self.initial_covariance = initial_covariance

        if transition_matrix is None:
            transition_matrix = jnp.eye(self.dim)
        self.transition_matrix = transition_matrix

        if transition_covariance is None:
            transition_covariance = jnp.eye(self.dim)
        self.transition_covariance = transition_covariance

        if likelihood_covariance_diag is None:
            likelihood_covariance_diag = jnp.eye(self.dim)
        self.likelihood_covariance_diag = likelihood_covariance_diag

        super().__init__(name=name)

    def __setattr__(self,
                    key: str,
                    value: Any):
        self.__dict__[key] = value
        if key[-10:] == 'covariance' or key[-15:] == 'covariance_diag':
            reset_covariance(self, key, value)

    def initial_potential(self,
                          x: jnp.ndarray,
                          t: Union[float, None]) -> Union[float, jnp.ndarray]:
        init_mean = self.initial_mean
        init_prec_sqrt = self.initial_precision_sqrt
        return gaussian_potential(x, init_mean, sqrt_prec=init_prec_sqrt)

    def initial_sample(self,
                       t: Union[float, None],
                       random_key: jnp.ndarray) -> Union[float, jnp.ndarray]:
        init_mean = self.initial_mean
        init_cov_sqrt = self.initial_covariance_sqrt
        return random.normal(random_key, shape=(self.dim,)) @ init_cov_sqrt.T + init_mean

    def transition_potential(self,
                             x_previous: jnp.ndarray,
                             t_previous: float,
                             x_new: jnp.ndarray,
                             t_new: float) -> Union[float, jnp.ndarray]:
        return gaussian_potential(x_new,
                                  x_previous @ self.transition_matrix.T,
                                  sqrt_prec=self.transition_precision_sqrt * jnp.sqrt(t_new - t_previous))

    def transition_sample(self,
                          x_previous: jnp.ndarray,
                          t_previous: float,
                          t_new: float,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        return x_previous @ self.transition_matrix.T \
               + random.normal(random_key, shape=x_previous.shape) @ self.transition_covariance_sqrt.T \
               * jnp.sqrt(t_new - t_previous)

    def likelihood_potential(self,
                             x: jnp.ndarray,
                             y: jnp.ndarray,
                             t: float) -> Union[float, jnp.ndarray]:
        return gaussian_potential(y,
                                  0,
                                  sqrt_prec=self.likelihood_precision_diag_sqrt / jnp.exp(0.5*x))

    def likelihood_sample(self,
                          x: jnp.ndarray,
                          t: float,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        return self.likelihood_covariance_diag_sqrt * jnp.exp(0.5*x) * random.normal(random_key, shape=x.shape)



########################################################################################################################
# Module: run.py
# Description: Sampler class and run function to generate samples
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import copy
from typing import Tuple, Union, Type
from functools import partial
from time import time
from inspect import isclass

from jax import numpy as jnp, jit
from mocat.src.core import cdict, static_cdict, Scenario
from mocat.utils import while_loop_stacked


class Sampler:
    parameters: cdict
    name: str
    max_iter: int = 10000

    def __init__(self,
                 name: str = None,
                 **kwargs):

        if name is not None:
            self.name = name

        if not hasattr(self, 'parameters'):
            self.parameters = cdict()

        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            else:
                setattr(self.parameters, key, value)

    def __repr__(self):
        return f"mocat.Sampler.{self.__class__.__name__}"

    def deepcopy(self) -> 'Sampler':
        return copy.deepcopy(self)

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: Union[None, cdict],
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        for key, value in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, value)
            if hasattr(self, 'parameters') and hasattr(self.parameters, key):
                setattr(self.parameters, key, value)

        if not hasattr(self, 'max_iter')\
            or not (isinstance(self.max_iter, int)
                    or (isinstance(self.max_iter, jnp.ndarray) and self.max_iter.dtype == 'int32')):
            raise AttributeError(self.__repr__() + ' max_iter must be int')

        if not hasattr(initial_extra, 'iter'):
            initial_extra.iter = 0

        if hasattr(self, 'parameters'):
            if not hasattr(initial_extra, 'parameters'):
                initial_extra.parameters = cdict()
            for key, value in self.parameters.__dict__.items():
                if not hasattr(initial_extra.parameters, key) or getattr(initial_extra.parameters, key) is None:
                    setattr(initial_extra.parameters, key, value)
        return initial_state, initial_extra

    def update(self,
               scenario: Scenario,
               state: cdict,
               extra: cdict) -> Tuple[cdict, cdict]:
        raise NotImplementedError(f'{self.name} update not initiated')

    def termination_criterion(self,
                              state: cdict,
                              extra: cdict) -> bool:
        return extra.iter >= self.max_iter

    def clean_chain(self,
                    scenario: Scenario,
                    chain_state: cdict):
        return chain_state

    def summary(self,
                scenario: Scenario,
                initial_state: cdict,
                initial_extra: cdict) -> cdict:
        summ = static_cdict()
        if hasattr(self, 'name'):
            summ.sampler = self.name

        if hasattr(scenario, 'name'):
            summ.scenario = scenario.name

        if hasattr(self, 'parameters'):
            summ.parameters = self.parameters

        if hasattr(self, 'tuning'):
            summ.tuning = self.tuning

        return summ


def run(scenario: Scenario,
        sampler: Union[Sampler, Type[Sampler]],
        n: int,
        random_key: Union[None, jnp.ndarray],
        initial_state: cdict = None,
        initial_extra: cdict = None,
        **kwargs) -> Union[cdict, Tuple[cdict, jnp.ndarray]]:

    if isclass(sampler):
        sampler = sampler(**kwargs)

    sampler.n = n

    if initial_extra is None:
        initial_extra = cdict()
    if random_key is not None:
        initial_extra.random_key = random_key

    initial_state, initial_extra = sampler.startup(scenario, n, initial_state, initial_extra,
                                                   **kwargs)

    summary = sampler.summary(scenario, initial_state, initial_extra)

    transport_kernel = partial(sampler.update, scenario)

    start = time()
    chain = while_loop_stacked(lambda state, extra: ~sampler.termination_criterion(state, extra),
                               transport_kernel,
                               (initial_state, initial_extra),
                               sampler.max_iter)
    chain = initial_state[jnp.newaxis] + chain
    chain = sampler.clean_chain(scenario, chain)
    chain.value.block_until_ready()
    end = time()
    chain.time = end - start

    chain.summary = summary

    return chain



# Analyse sensitivity of maximum_rejections parameter (and N) for online smoothing

import matplotlib.pyplot as plt
from jax import numpy as jnp, random, jit
from scipy.interpolate import make_interp_spline
import os
from time import time
import pickle

import mocat
from mocat import ssm

fig_dir = os.path.dirname(os.getcwd()) + '/images/linear_gaussian_1D'

# Use splines to smooth trajectories
def smooth(t, x, k=3):
    t_new = jnp.linspace(t.min(), t.max(), 200)
    spl = make_interp_spline(t, x, k=k)
    y_smooth = spl(t_new)
    return t_new, y_smooth


# Initiate random seed
random_keys = random.split(random.PRNGKey(3), 100)

# Number of particles to generate
ns = [200, 400, 600, 800, 1000]

# Transition variance
sigma2_x = 1.

# Observation variance
sigma2_y = 1.

# Lag
lag = 5

# Rejections
max_rejs = [0, 1, 2, 4, 8, 16, 32]

# Initiate scenario
dim = 1
lgssm = ssm.TimeHomogenousLinearGaussian(initial_mean=jnp.zeros(dim),
                                         initial_covariance=jnp.eye(dim) * sigma2_x,
                                         transition_matrix=jnp.eye(dim),
                                         transition_covariance=jnp.eye(dim) * sigma2_x,
                                         likelihood_matrix=jnp.eye(dim),
                                         likelihood_covariance=jnp.eye(dim) * sigma2_y,
                                         name='1D LGSSM')

# Simulate underlying true hidden process and observations
len_t = 40
true_process = lgssm.simulate(jnp.arange(len_t, dtype='float32'), random_keys[0])

# Initiate particle filter
pf = ssm.OptimalNonLinearGaussianParticleFilter()

# Online smoother pf fixed lag
pf_fl_rk = random_keys[5]
pf_fl_samps = []
for n_int in ns:
    pf_fl_samps_int = []
    for r in max_rejs:
        pf_fl_rk, _ = random.split(pf_fl_rk)
        pf_fl_rks = random.split(pf_fl_rk, len_t)
        print(f'OPS - wPF - N = {n_int}')
        propagate_fl_pf = lambda i, parts: ssm.propagate_particle_smoother(lgssm, pf, parts, true_process.y[i],
                                                                           true_process.t[i], pf_fl_rks[i], lag, False,
                                                                           maximum_rejections=r)
        propagate_fl_pf = jit(propagate_fl_pf)
        start = time()
        pf_fl_trajectories = ssm.initiate_particles(lgssm, pf, n_int, pf_fl_rks[0], true_process.y[0], true_process.t[0])
        for i in range(1, len_t):
            pf_fl_trajectories = propagate_fl_pf(i, pf_fl_trajectories)
        pf_fl_trajectories.value.block_until_ready()
        pf_fl_trajectories.time = time() - start
        print(pf_fl_trajectories.time)
        pf_fl_trajectories.lag = lag
        pf_fl_trajectories.maximum_rejections = max_rejs
        pf_fl_samps_int.append(pf_fl_trajectories.copy())
    pf_fl_samps.append(pf_fl_samps_int)

with open(fig_dir + '/fl_pf_varynr.pkl', 'wb') as f:
    pickle.dump(pf_fl_samps, f)

# Online smoother bs fixed lag
bs_fl_rk = random_keys[6]
bs_fl_samps = []
for n_int in ns:
    bs_fl_samps_int = []
    for r in max_rejs:
        bs_fl_rk, _ = random.split(bs_fl_rk)
        bs_fl_rks = random.split(bs_fl_rk, len_t)
        print(f'OPS - wBS - N = {n_int} - R = {r}')
        propagate_fl_bs = lambda i, parts: ssm.propagate_particle_smoother(lgssm, pf, parts, true_process.y[i],
                                                                           true_process.t[i], bs_fl_rks[i], lag, True,
                                                                           maximum_rejections=r)
        propagate_fl_bs = jit(propagate_fl_bs)
        start = time()
        bs_fl_trajectories = ssm.initiate_particles(lgssm, pf, n_int, bs_fl_rks[0], true_process.y[0], true_process.t[0])
        for i in range(1, len_t):
            bs_fl_trajectories = propagate_fl_bs(i, bs_fl_trajectories)
        bs_fl_trajectories.value.block_until_ready()
        bs_fl_trajectories.time = time() - start
        print(bs_fl_trajectories.time)
        bs_fl_trajectories.lag = lag
        bs_fl_trajectories.maximum_rejections = max_rejs
        bs_fl_samps_int.append(bs_fl_trajectories.copy())
    bs_fl_samps.append(bs_fl_samps_int)

with open(fig_dir + '/fl_bs_varynr.pkl', 'wb') as f:
    pickle.dump(bs_fl_samps, f)

# Plot num_transition_evals
pf_fl_nte = jnp.array(
    [[pf_fl_samps[j][i].num_transition_evals.mean() for i in range(len(max_rejs))] for j in range(len(ns))]) / 1e6
bs_fl_nte = jnp.array(
    [[bs_fl_samps[j][i].num_transition_evals.mean() for i in range(len(max_rejs))] for j in range(len(ns))]) / 1e6

line_styles = ['-', '--', ':', '-.', (0, (10, 10))]
ymax_pf = 0.4
fig_pf, ax_pf = plt.subplots()
for i in range(len(ns)):
    ax_pf.plot(max_rejs, pf_fl_nte[i], color='red', label=f'{ns[i]}', linestyle=line_styles[i])
ax_pf.set_xlabel(r'$R$')
ax_pf.set_ylabel(r'Number of transition density evaluations (millions)')
ax_pf.set_ylim(0, ymax_pf)
handles, labels = ax_pf.get_legend_handles_labels()
ax_pf.legend(handles[::-1], labels[::-1], title=r'$N$')
fig_pf.savefig(fig_dir + f'/pf_fl_num_transition_evals_nr', dpi=300)

fig_bs, ax_bs = plt.subplots()
ymax_bs = 2.0
for i in range(len(ns)):
    ax_bs.plot(max_rejs, bs_fl_nte[i], color='blue', label=f'{ns[i]}', linestyle=line_styles[i])
ax_bs.set_xlabel(r'$R$')
ax_bs.set_ylabel(r'Number of transition density evaluations (millions)')
ax_bs.set_ylim(0, ymax_bs)
handles, labels = ax_bs.get_legend_handles_labels()
ax_bs.legend(handles[::-1], labels[::-1], title=r'$N$')
fig_bs.savefig(fig_dir + f'/bs_fl_num_transition_evals_nr', dpi=300)





########################################################################################################################
# Module: tests/test_mcmc_sampling.py
# Description: Tests for MCMC sampling.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

import unittest

import jax.numpy as jnp
from jax.random import PRNGKey
import numpy.testing as npt

from mocat.src.core import cdict
from mocat.src.scenarios.twodim import toy_examples
from mocat.src.metrics import ess_autocorrelation
from mocat.src.mcmc import standard_mcmc
from mocat.src.mcmc.metropolis import Metropolis, RMMetropolis
from mocat.src.sample import run


class TestMetropolisCorrelatedGaussian(unittest.TestCase):

    sampler = None

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)

        self.scenario_cov = jnp.array([[1., 0.9], [0.9, 2.]])
        self.scenario = toy_examples.Gaussian(covariance=self.scenario_cov)
        self.n = int(1e5)

        self.__test__ = self.sampler is not None
        if self.__test__:
            self.adapt_sample = run(self.scenario, self.sampler, self.n, PRNGKey(0), correction=RMMetropolis())
            self.sampler.parameters.stepsize = self.adapt_sample.stepsize[-1]
            self.warmstart_sample = run(self.scenario, self.sampler, self.n, PRNGKey(0), correction=Metropolis)

    def _test_mean(self,
                   sample: cdict):
        if sample.value.ndim == 3:
            val = jnp.concatenate(sample.value)
        else:
            val = sample.value
        samp_mean = val.mean(axis=0)
        npt.assert_array_almost_equal(samp_mean, jnp.zeros(2), decimal=1)

    def _test_cov(self,
                  sample: cdict):
        if sample.value.ndim == 3:
            val = jnp.concatenate(sample.value)
        else:
            val = sample.value
        samp_cov = jnp.cov(val.T)
        npt.assert_array_almost_equal(samp_cov, self.scenario_cov, decimal=1)

    def _test_ess_autocorrelation(self,
                  sample: cdict):
        samp_ess_pot = ess_autocorrelation(sample)
        samp_ess_0 = ess_autocorrelation(sample.value[:, 0])
        samp_ess_1 = ess_autocorrelation(sample.value[:, 1])

        self.assertFalse(samp_ess_pot == samp_ess_0)
        self.assertFalse(samp_ess_pot == samp_ess_1)
        self.assertFalse(samp_ess_0 == samp_ess_1)

        self.assertGreater(samp_ess_pot, self.n / 100)
        self.assertGreater(samp_ess_0, self.n / 100)
        self.assertGreater(samp_ess_1, self.n / 100)

    def _test_acceptance_rate(self,
                              sample: cdict):
        npt.assert_almost_equal(sample.alpha.mean(), self.sampler.tuning.target, decimal=1)


class TestRandomWalkCorrelatedGaussian(TestMetropolisCorrelatedGaussian):
    sampler = standard_mcmc.RandomWalk()

    def test_mean(self):
        self._test_mean(self.adapt_sample)
        self._test_mean(self.warmstart_sample)

    def test_cov(self):
        self._test_cov(self.adapt_sample)
        self._test_cov(self.warmstart_sample)

    def test_acceptance_rate(self):
        self._test_acceptance_rate(self.adapt_sample)
        self._test_acceptance_rate(self.warmstart_sample)

    def test_ess_autocorrelation(self):
        self._test_ess_autocorrelation(self.adapt_sample)
        self._test_ess_autocorrelation(self.warmstart_sample)


class TestUnderdampedCorrelatedGaussian(TestMetropolisCorrelatedGaussian):
    sampler = standard_mcmc.Underdamped(leapfrog_steps=3,
                                        friction=1.)

    def test_mean(self):
        self._test_mean(self.adapt_sample)
        self._test_mean(self.warmstart_sample)

    def test_cov(self):
        self._test_cov(self.adapt_sample)
        self._test_cov(self.warmstart_sample)

    def test_acceptance_rate(self):
        self._test_acceptance_rate(self.warmstart_sample)

    def test_ess_autocorrelation(self):
        self._test_ess_autocorrelation(self.adapt_sample)
        self._test_ess_autocorrelation(self.warmstart_sample)


if __name__ == '__main__':
    unittest.main()




# Visualise a variety of smoothing techniques for a one-dimensional LGSSM

import matplotlib.pyplot as plt
from jax import numpy as jnp, random, jit
from scipy.interpolate import make_interp_spline
import os
from time import time

from mocat import ssm

fig_dir = os.path.dirname(os.getcwd()) + '/images/linear_gaussian_1D'

# Plot params
particle_scatter_size = 0.8
particle_line_width = 0.4
particle_colour = 'orange'
particle_alpha = 0.5
particle_zorder = 0
truth_line_width = 2.
truth_colour = 'red'
truth_zorder = 3
posterior_colour = 'blue'
posterior_alpha = 0.25
ylim = (-8, 1)


# Use splines to smooth trajectories
def smooth(t, x, k=3):
    t_new = jnp.linspace(t.min(), t.max(), 200)
    spl = make_interp_spline(t, x, k=k)
    y_smooth = spl(t_new)
    return t_new, y_smooth


# Initiate random seed
random_keys = random.split(random.PRNGKey(3), 100)

# Number of particles to generate
n = 50

# Transition variance
sigma2_x = 1.

# Observation variance
sigma2_y = 1.

# Lags
lags = [2, 10]

# Rejections
max_rej = 0

# Initiate scenario
dim = 1
lgssm = ssm.TimeHomogenousLinearGaussian(initial_mean=jnp.zeros(dim),
                                         initial_covariance=jnp.eye(dim) * sigma2_x,
                                         transition_matrix=jnp.eye(dim),
                                         transition_covariance=jnp.eye(dim) * sigma2_x,
                                         likelihood_matrix=jnp.eye(dim),
                                         likelihood_covariance=jnp.eye(dim) * sigma2_y,
                                         name='1D LGSSM')

# Simulate underlying true hidden process and observations
len_t = 40
true_process = lgssm.simulate(jnp.arange(len_t, dtype='float32'), random_keys[0])

# Initiate particle filter
pf = ssm.OptimalNonLinearGaussianParticleFilter()

# Plot truth and observations
fig, ax = plt.subplots()
ax.plot(*smooth(true_process.t, true_process.x[:, 0]),
        c=truth_colour, zorder=truth_zorder, linewidth=truth_line_width)
ax.scatter(true_process.t, true_process.y[:, 0], c=truth_colour, marker='x', zorder=truth_zorder)
ax.set_xlabel(r'$t$')
ax.set_ylim(ylim)
fig.savefig(fig_dir + '/truth', dpi=300)

# Analytical posterior
mean_x = jnp.zeros(len_t)
mean_y = jnp.zeros(len_t)
cov_x = jnp.array([[min(i, j) + 1 for i in range(len_t)] for j in range(len_t)], dtype='float32')
prec_x = jnp.linalg.inv(cov_x)
cov_y = cov_x + jnp.eye(len_t)
prec_y = jnp.linalg.inv(cov_y)
mean_post = mean_x + cov_x @ prec_y @ (true_process.y[:, 0] - mean_y)
cov_post = cov_x - cov_x @ prec_y @ cov_x
prec_post = jnp.linalg.inv(cov_post)
fig, ax = plt.subplots()
smoothed_mean_post = smooth(true_process.t, mean_post)
ax.plot(*smoothed_mean_post, c=posterior_colour, zorder=particle_zorder, linewidth=truth_line_width)
margsds = jnp.sqrt(jnp.diag(cov_post))
telong, smooth_lower = smooth(true_process.t, mean_post - margsds)
smooth_upper = smooth(true_process.t, mean_post + margsds)[1]
ax.fill_between(telong, smooth_lower, smooth_upper,
                color=posterior_colour, alpha=posterior_alpha, zorder=particle_zorder)
ax.plot(*smooth(true_process.t, true_process.x[:, 0]),
        c=truth_colour, zorder=truth_zorder, linewidth=truth_line_width)
ax.scatter(true_process.t, true_process.y[:, 0], c=truth_colour, marker='x', zorder=truth_zorder)
ax.set_xlabel(r'$t$')
ax.set_ylim(ylim)
fig.savefig(fig_dir + '/posterior', dpi=300)

# Run particle filter for marginals
pf_marginals = ssm.run_particle_filter_for_marginals(lgssm, pf, true_process.y, true_process.t, random_keys[1], n,
                                                     ess_threshold=1.)
fig, ax = plt.subplots()
for i in range(n):
    ax.scatter(true_process.t, pf_marginals.value[:, i, 0], marker='o', c=particle_colour,
               alpha=particle_alpha, s=particle_scatter_size, zorder=particle_zorder)
ax.plot(*smoothed_mean_post, c=posterior_colour, zorder=truth_zorder, linewidth=truth_line_width)
ax.scatter(true_process.t, true_process.y[:, 0], c=truth_colour, marker='x', zorder=truth_zorder)
ax.set_xlabel(r'$t$')
ax.set_ylim(ylim)
fig.savefig(fig_dir + '/pf_marginals', dpi=300)

# Run particle filter for trajectories
pft_rks = random.split(random_keys[2], len_t)
pf_trajectories = ssm.initiate_particles(lgssm, pf, n, pft_rks[0], true_process.y[0], true_process.t[0])
for i in range(1, len_t):
    pf_trajectories = ssm.propagate_particle_filter(lgssm, pf, pf_trajectories, true_process.y[i], true_process.t[i],
                                                    pft_rks[i], ess_threshold=1.)
fig, ax = plt.subplots()
for i in range(n):
    ax.plot(*smooth(true_process.t, pf_trajectories.value[:, i, 0]), c=particle_colour,
            alpha=particle_alpha, linewidth=particle_line_width, zorder=particle_zorder)
ax.plot(*smoothed_mean_post, c=posterior_colour, zorder=truth_zorder, linewidth=truth_line_width)
ax.scatter(true_process.t, true_process.y[:, 0], c=truth_colour, marker='x', zorder=truth_zorder)
ax.set_xlabel(r'$t$')
ax.set_ylim(ylim)
fig.savefig(fig_dir + '/pf_joint', dpi=300)

# Backward simulation
start = time()
backsim = ssm.backward_simulation(lgssm, pf_marginals, random_keys[3], n)
backsim.value.block_until_ready()
backsim.time = time() - start
fig, ax = plt.subplots()
for i in range(n):
    ax.plot(*smooth(true_process.t, backsim.value[:, i, 0]), c=particle_colour,
            alpha=particle_alpha, linewidth=particle_line_width, zorder=particle_zorder)
ax.plot(*smoothed_mean_post, c=posterior_colour, zorder=truth_zorder, linewidth=truth_line_width)
ax.scatter(true_process.t, true_process.y[:, 0], c=truth_colour, marker='x', zorder=truth_zorder)
ax.set_xlabel(r'$t$')
ax.set_ylim(ylim)
fig.savefig(fig_dir + '/ffbsi', dpi=300)

# Marginal fixed lag
mfl_rk = random_keys[4]
for lag in lags:
    mfl_rk, _ = random.split(mfl_rk)
    mfl_rks = random.split(mfl_rk, len_t)
    vals = []
    mfl_trajectories = ssm.initiate_particles(lgssm, pf, n, mfl_rks[0], true_process.y[0], true_process.t[0])
    for i in range(1, len_t):
        mfl_trajectories = ssm.propagate_particle_filter(lgssm, pf, mfl_trajectories, true_process.y[i],
                                                         true_process.t[i], mfl_rks[i], ess_threshold=1.)
        if i - lag >= 0:
            vals.append(mfl_trajectories.value[i - lag].copy())

    marg_vals = jnp.array(vals)
    joint_vals = mfl_trajectories.value[(len_t - lag):]
    fig, ax = plt.subplots()
    for i in range(n):
        ax.scatter(true_process.t[:(len_t - lag)], marg_vals[:, i, 0], marker='o', c=particle_colour,
                   alpha=particle_alpha, s=particle_scatter_size, zorder=particle_zorder)
    for i in range(n):
        ax.plot(*smooth(true_process.t[(len_t - lag):], joint_vals[:, i, 0], k=min(3, lag - 1)), c=particle_colour,
                alpha=particle_alpha, linewidth=particle_line_width, zorder=particle_zorder)
    ax.plot(*smoothed_mean_post, c=posterior_colour, zorder=truth_zorder, linewidth=truth_line_width)
    ax.scatter(true_process.t, true_process.y[:, 0], c=truth_colour, marker='x', zorder=truth_zorder)
    ax.set_xlabel(r'$t$')
    ax.set_ylim(ylim)
    fig.savefig(fig_dir + f'/marginal_fixed_lag={lag}', dpi=300)

# Online smoother pf fixed lag
pf_fl_rk = random_keys[5]
for lag in lags:
    pf_fl_rk, _ = random.split(pf_fl_rk)
    pf_fl_rks = random.split(pf_fl_rk, len_t)
    print(f'OPS - wPF - lag = {lag}')
    propagate_fl_pf = lambda i, parts: ssm.propagate_particle_smoother(lgssm, pf, parts, true_process.y[i],
                                                                       true_process.t[i], pf_fl_rks[i], lag, False,
                                                                       maximum_rejections=max_rej)
    propagate_fl_pf = jit(propagate_fl_pf)
    start = time()
    pf_fl_trajectories = ssm.initiate_particles(lgssm, pf, n, pf_fl_rks[0], true_process.y[0], true_process.t[0])
    for i in range(1, len_t):
        pf_fl_trajectories = propagate_fl_pf(i, pf_fl_trajectories)
    pf_fl_trajectories.value.block_until_ready()
    pf_fl_trajectories.time = time() - start
    print(pf_fl_trajectories.time)

    fig, ax = plt.subplots()
    for i in range(n):
        ax.plot(*smooth(true_process.t, pf_fl_trajectories.value[:, i, 0]), c=particle_colour,
                alpha=particle_alpha, linewidth=particle_line_width, zorder=particle_zorder)
    ax.plot(*smoothed_mean_post, c=posterior_colour, zorder=truth_zorder, linewidth=truth_line_width)
    ax.scatter(true_process.t, true_process.y[:, 0], c=truth_colour, marker='x', zorder=truth_zorder)
    ax.set_xlabel(r'$t$')
    ax.set_ylim(ylim)
    fig.savefig(fig_dir + f'/online_smooth_pf_lag={lag}', dpi=300)

# Online smoother bs fixed lag
bs_fl_rk = random_keys[6]
for lag in lags:
    bs_fl_rk, _ = random.split(bs_fl_rk)
    bs_fl_rks = random.split(bs_fl_rk, len_t)
    print(f'OPS - wBS - lag = {lag}')
    propagate_fl_bs = lambda i, parts: ssm.propagate_particle_smoother(lgssm, pf, parts, true_process.y[i],
                                                                       true_process.t[i], bs_fl_rks[i], lag, True,
                                                                       maximum_rejections=max_rej, ess_threshold=1.)
    propagate_fl_bs = jit(propagate_fl_bs)
    start = time()
    bs_fl_trajectories = ssm.initiate_particles(lgssm, pf, n, bs_fl_rks[0], true_process.y[0], true_process.t[0])
    for i in range(1, len_t):
        bs_fl_trajectories = propagate_fl_bs(i, bs_fl_trajectories)
    bs_fl_trajectories.value.block_until_ready()
    bs_fl_trajectories.time = time() - start
    print(bs_fl_trajectories.time)

    fig, ax = plt.subplots()
    for i in range(n):
        ax.plot(*smooth(true_process.t, bs_fl_trajectories.value[:, i, 0]), c=particle_colour,
                alpha=particle_alpha, linewidth=particle_line_width, zorder=particle_zorder)
    ax.plot(*smoothed_mean_post, c=posterior_colour, zorder=truth_zorder, linewidth=truth_line_width)
    ax.scatter(true_process.t, true_process.y[:, 0], c=truth_colour, marker='x', zorder=truth_zorder)
    ax.set_xlabel(r'$t$')
    ax.set_ylim(ylim)
    fig.savefig(fig_dir + f'/online_smooth_bs_lag={lag}', dpi=300)



########################################################################################################################
# Module: transport/sampler.py
# Description: Generic transport sampler class - iteratively moves an ensemble of particles.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple

from jax import random, vmap

from mocat.src.core import Scenario, cdict, is_implemented
from mocat.src.sample import Sampler


class TransportSampler(Sampler):

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        if initial_state is None:
            initial_extra.random_key, sub_key = random.split(initial_extra.random_key)
            if is_implemented(scenario.prior_sample):
                init_vals = vmap(scenario.prior_sample)(random.split(sub_key, n))
            else:
                init_vals = random.normal(sub_key, shape=(n, scenario.dim))
            initial_state = cdict(value=init_vals)

        initial_state, initial_extra = super().startup(scenario, n, initial_state, initial_extra, **kwargs)

        return initial_state, initial_extra

    def update(self,
               scenario: Scenario,
               ensemble_state: cdict,
               extra: cdict) -> Tuple[cdict, cdict]:
        raise NotImplementedError(f'{self.name} update not initiated')

    def clean_chain(self,
                    scenario: Scenario,
                    chain_ensemble_state: cdict) -> cdict:
        return chain_ensemble_state

    def termination_criterion(self,
                              ensemble_state: cdict,
                              extra: cdict) -> bool:
        return extra.iter >= self.max_iter




########################################################################################################################
# Module: metrics.py
# Description: Some common metrics to assess run quality.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union, Callable, Tuple
from warnings import warn
from functools import partial

from decorator import decorator
from jax import vmap, random, numpy as jnp
from jax.scipy.special import logsumexp
from matplotlib import pyplot as plt
from matplotlib.figure import Figure

from mocat.src.core import cdict
from mocat.src.kernels import Kernel


def extract_1d_vals(sample: Union[jnp.ndarray, cdict]) -> jnp.ndarray:
    vals = sample.potential if isinstance(sample, cdict) else sample
    if vals.ndim > 1:
        raise TypeError('vals must be 1 dimensional array or cdict with 1 dimensional potential')
    return vals


def _next_pow_two(n: int) -> int:
    i = 1
    while i < n:
        i = i << 1
    return i


def autocorrelation(sample: Union[jnp.ndarray, cdict],
                    max_lag_eval: int = 1000) -> jnp.ndarray:
    vals = extract_1d_vals(sample)
    max_lag_eval = min(max_lag_eval, len(vals))
    n = _next_pow_two(max_lag_eval)
    f = jnp.fft.fft(vals - jnp.mean(vals), n=2 * n)
    acf = jnp.fft.ifft(f * jnp.conjugate(f))[:max_lag_eval].real
    return acf / acf[0]


def integrated_autocorrelation_time(sample: Union[jnp.ndarray, cdict],
                                    max_lag_iat: int = None,
                                    max_lag_eval: int = 1000) -> float:
    autocorr = autocorrelation(sample, max_lag_eval)
    iats_all = 1 + 2 * jnp.cumsum(autocorr[1:])

    # https://dfm.io/posts/autocorr/
    if max_lag_iat is None:
        c = 5
        m_minus_c_tau = jnp.arange(1, len(iats_all) + 1) - c * iats_all
        max_lag_iat = jnp.argmax(m_minus_c_tau > 0)

    return iats_all[max_lag_iat]


def ess_autocorrelation(sample: Union[jnp.ndarray, cdict],
                        max_lag_iat: int = None,
                        max_lag_eval: int = 1000) -> float:
    vals = extract_1d_vals(sample)
    iat = integrated_autocorrelation_time(vals, max_lag_iat, max_lag_eval)
    return len(vals) / iat


def log_ess_log_weight(log_weight: Union[jnp.ndarray, cdict]) -> float:
    if isinstance(log_weight, cdict) and hasattr(log_weight, 'log_weight'):
        log_weight = log_weight.log_weight
    if not isinstance(log_weight, jnp.ndarray):
        raise TypeError('log_weight must be jnp.ndarray or cdict with log_weight attribute')
    return 2 * logsumexp(log_weight) - logsumexp(2 * log_weight)


def ess_log_weight(log_weight: Union[jnp.ndarray, cdict]) -> float:
    return jnp.exp(log_ess_log_weight(log_weight))


def squared_jumping_distance(sample: Union[jnp.ndarray, cdict]) -> float:
    vals = sample.value if isinstance(sample, cdict) else sample
    sjd = vals[1:] - vals[:-1]
    sjd = sjd ** 2
    return sjd.mean(0)


def ksd(sample: Union[jnp.ndarray, cdict],
        kernel: Kernel,
        grad_potential: jnp.ndarray = None,
        log_weight: jnp.ndarray = None,
        ensemble_batchsize: int = None,
        random_key: jnp.ndarray = None,
        **kernel_params) -> float:
    vals = sample.value if isinstance(sample, cdict) else sample
    n = len(vals)

    if grad_potential is None and isinstance(sample, cdict) and hasattr(sample, 'grad_potential'):
        grad_potential = sample.grad_potential
    elif grad_potential is None:
        raise TypeError('grad_potential not found')

    if ensemble_batchsize is None:
        get_batch_inds = lambda _: jnp.arange(n)
    else:
        random_inds = random.choice(random_key, n, shape=(n, ensemble_batchsize))
        get_batch_inds = lambda i: random_inds[i]

    if log_weight is not None:
        weights = jnp.exp(log_weight)
    else:
        weights = jnp.ones(n)

    sum_weights = weights.sum()

    def k_0_inds(x_i: int, y_i: int) -> float:
        kern_diag_grad_xy_mat = jnp.sum(kernel.diag_grad_xy(vals[x_i], vals[y_i], **kernel_params))
        grad_kx_grad_py_mat = jnp.dot(kernel.grad_x(vals[x_i], vals[y_i], **kernel_params), grad_potential[y_i])
        grad_ky_grad_px_mat = jnp.dot(grad_potential[x_i], kernel.grad_y(vals[x_i], vals[y_i], **kernel_params))
        kern_grad_pxy = kernel(vals[x_i], vals[y_i], **kernel_params) * jnp.dot(grad_potential[x_i],
                                                                                grad_potential[y_i])

        return (kern_diag_grad_xy_mat + grad_kx_grad_py_mat + grad_ky_grad_px_mat + kern_grad_pxy) \
               * weights[x_i] * weights[y_i]

    def v_k_0(i: int) -> jnp.ndarray:
        batch_inds = get_batch_inds(i)
        return vmap(partial(k_0_inds, i))(batch_inds) / (sum_weights * weights[batch_inds].sum())

    return jnp.sqrt((vmap(v_k_0)(jnp.arange(n))).sum())


@decorator
def metric_plot(plot_func: Callable,
                *args,
                **kwargs) -> Union[plt.Axes, Tuple[Figure, plt.Axes]]:
    ax = args[-1]
    if ax is None:
        fig, ax = plt.subplots()
        plot_func(*args[:-1], ax, **kwargs)
        return fig, ax
    else:
        plot_func(*args, **kwargs)
        return ax


@metric_plot
def autocorrelation_plot(sample: Union[jnp.ndarray, cdict],
                         max_lag_plot: int = 100,
                         max_lag_eval: int = 1000,
                         ax: plt.Axes = None,
                         **kwargs):
    vals = extract_1d_vals(sample)

    autocor = autocorrelation(vals, max_lag_eval)
    ax.bar(range(1, 1 + len(autocor)), autocor, **kwargs)
    ax.set_xlim(0, max_lag_plot)


@metric_plot
def trace_plot(sample: Union[jnp.ndarray, cdict],
               last_n: int = None,
               ax: plt.Axes = None,
               **kwargs):
    vals = extract_1d_vals(sample)

    if last_n is None:
        start = 0
    else:
        start = vals.shape[0] - last_n
    ax.plot(vals[start:], **kwargs)


@metric_plot
def plot_2d_samples(sample: Union[jnp.ndarray, cdict],
                    dim1: int = 0,
                    dim2: int = 1,
                    s: float = 0.5,
                    ax: plt.Axes = None,
                    **kwargs):
    vals = sample.value if isinstance(sample, cdict) else sample

    if vals.ndim == 3:
        vals = jnp.concatenate(vals)
        warn('Concatenating 3 dimensional array to plot')
    if vals.ndim == 1 or vals.shape[-1] == 1:
        raise TypeError('Samples are one dimensional - try mocat.hist_1d_samples')
    ax.scatter(vals[:, dim1], vals[:, dim2], s, **kwargs)


@metric_plot
def hist_1d_samples(sample: Union[jnp.ndarray, cdict],
                    dim: int = 0,
                    bins: int = 50,
                    density: bool = True,
                    ax: plt.Axes = None,
                    **kwargs):
    vals = sample.value if isinstance(sample, cdict) else sample

    if vals.ndim == 1:
        vals = vals[..., jnp.newaxis]
    if vals.ndim == 3:
        vals = jnp.concatenate(vals)

    ax.hist(vals[:, dim], bins=bins, density=density, **kwargs)



########################################################################################################################
# Module: ssm/scenarios/nonlinear1d.py
# Description: One-dimensional non-linear benchmark model.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################


from typing import Union

from jax import numpy as jnp, random
from mocat.src.ssm.ssm import StateSpaceModel
from mocat.src.utils import gaussian_potential


class NonLinear1DBenchmark(StateSpaceModel):
    name = '1D Non-linear Benchmark'
    dim = 1
    dim_obs = 1

    def __init__(self,
                 initial_sd: float = jnp.sqrt(2.),
                 transition_sd: float = jnp.sqrt(10.),
                 likelihood_sd: float = 1.,
                 name: str = None):
        self.initial_sd = initial_sd
        self.transition_sd = transition_sd
        self.likelihood_sd = likelihood_sd
        super().__init__(name=name)

    def initial_potential(self,
                          x: jnp.ndarray,
                          t: float) -> Union[float, jnp.ndarray]:
        return gaussian_potential(x / self.initial_sd)

    def initial_sample(self,
                       t: float,
                       random_key: jnp.ndarray) -> Union[float, jnp.ndarray]:
        return random.normal(random_key, (1,)) * self.initial_sd

    def _transition_mean(self,
                         x_previous: jnp.ndarray,
                         t_previous: float) -> jnp.ndarray:
        return 0.5 * x_previous\
               + 25 * x_previous / (1 + x_previous ** 2)\
               + 8 * jnp.cos(1.2 * t_previous)

    def transition_potential(self,
                             x_previous: jnp.ndarray,
                             t_previous: float,
                             x_new: jnp.ndarray,
                             t_new: float) -> Union[float, jnp.ndarray]:
        transition_mean = self._transition_mean(x_previous, t_previous)
        return gaussian_potential((x_new - transition_mean) / self.transition_sd)

    def transition_sample(self,
                          x_previous: jnp.ndarray,
                          t_previous: float,
                          t_new: float,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        transition_mean = self._transition_mean(x_previous, t_previous)
        return transition_mean + random.normal(random_key, (1,)) * self.transition_sd

    def _likelihood_mean(self,
                         x: jnp.ndarray,
                         t: float) -> jnp.ndarray:
        return x ** 2 / 20

    def likelihood_potential(self,
                             x: jnp.ndarray,
                             y: jnp.ndarray,
                             t: float) -> Union[float, jnp.ndarray]:
        lik_mean = self._likelihood_mean(x, t)
        return gaussian_potential((y - lik_mean)/self.likelihood_sd)

    def likelihood_sample(self,
                          x: jnp.ndarray,
                          t: float,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        lik_mean = self._likelihood_mean(x, t)
        return lik_mean + random.normal(random_key, (1,)) * self.likelihood_sd


########################################################################################################################
# Module: abc/scenarios/gk.py
# Description: Stochastic Lotka-Volterra (predator-prey).
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union, Tuple

from jax import numpy as jnp, random

from mocat.src.abc.abc import ABCScenario
from mocat.src.utils import _while_loop_stacked

exponential_sample_buffer = 1e-6


def lotka_volterra_single_step(carry: Tuple[jnp.ndarray, float], extra: Tuple[jnp.ndarray, jnp.ndarray]):
    prey_pred, current_time = carry
    params, random_key = extra
    t_key, haz_key, random_key = random.split(random_key, 3)

    hazards = jnp.array([params[0] * prey_pred[0],
                         params[1] * prey_pred[0] * prey_pred[1],
                         params[2] * prey_pred[1]])
    sum_hazards = hazards.sum()
    t_increment = -jnp.log(exponential_sample_buffer + random.uniform(t_key)) / sum_hazards
    hazard_probs = hazards / sum_hazards

    possibilities = jnp.array([[prey_pred[0] + 1, prey_pred[1]],
                               [prey_pred[0] - 1, prey_pred[1] + 1],
                               [prey_pred[0], prey_pred[1] - 1]])
    return (possibilities[random.choice(haz_key, 3, p=hazard_probs)], current_time + t_increment), (params, random_key)


def lotka_volterra_simulate(initial_prey_pred: jnp.ndarray,
                            times: jnp.ndarray,
                            params: jnp.ndarray,
                            random_key: jnp.ndarray,
                            max_iter: int = 10000) -> jnp.ndarray:
    max_time = jnp.max(times)

    (simulated_pred_prey, simulated_times), _, _ = _while_loop_stacked(lambda carry, extra: carry[1] < max_time,
                                                                       lotka_volterra_single_step,
                                                                       ((initial_prey_pred, times[0]),
                                                                        (params, random_key)),
                                                                       max_iter)

    simulated_times = jnp.where(simulated_times == 0, jnp.inf, simulated_times)
    return simulated_pred_prey[jnp.searchsorted(simulated_times, times[1:]) - 1]


class LotkaVolterra(ABCScenario):
    name: str = 'Lotka-Volterra'
    dim: int = 3

    initial_prey_pred: jnp.ndarray = None
    prior_rates: jnp.ndarray = jnp.ones(dim)
    times: jnp.ndarray = None
    data: jnp.ndarray = None
    max_iter: int = 10000

    def prior_sample(self,
                     random_key: jnp.ndarray) -> jnp.ndarray:
        return -jnp.log(exponential_sample_buffer + random.uniform(random_key, (self.dim,))) / self.prior_rates

    def prior_potential(self,
                        x: jnp.ndarray,
                        random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return (self.prior_rates * x).sum()

    def likelihood_sample(self,
                          x: jnp.ndarray,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        return lotka_volterra_simulate(self.initial_prey_pred, self.times, x, random_key, max_iter=self.max_iter)[:, 0]


class TransformedLotkaVolterra(LotkaVolterra):

    @staticmethod
    def constrain(x):
        return jnp.exp(x)

    @staticmethod
    def unconstrain(x):
        return jnp.log(x)

    def prior_sample(self,
                     random_key: jnp.ndarray) -> jnp.ndarray:
        return self.unconstrain(super().prior_sample(random_key))

    def prior_potential(self,
                        x: jnp.ndarray,
                        random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return (self.prior_rates * jnp.exp(x) - x).sum()

    def likelihood_sample(self,
                          x: jnp.ndarray,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        return super().likelihood_sample(self.constrain(x), random_key)



########################################################################################################################
# Module: ssm/linear_gaussian/linear_gaussian_1D_N500.py
# Description: Kalman filtering/smoothing for linear Gaussian state-space models.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple

from jax import numpy as jnp
from jax.lax import scan

from mocat.src.ssm.linear_gaussian.linear_gaussian import LinearGaussian


def run_kalman_filter_for_marginals(lgssm_scenario: LinearGaussian,
                                    y: jnp.ndarray,
                                    t: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
    mu_0 = lgssm_scenario.get_initial_mean(t[0])
    cov_0 = lgssm_scenario.get_initial_covariance_sqrt(t[0])

    l_mat_0 = lgssm_scenario.get_likelihood_matrix(t[0])
    l_cov_sqrt_0 = lgssm_scenario.get_likelihood_covariance_sqrt(t[0])
    l_cov_0 = l_cov_sqrt_0 @ l_cov_sqrt_0.T

    kal_0 = cov_0 @ l_mat_0.T @ jnp.linalg.inv(l_mat_0 @ cov_0 @ l_mat_0.T + l_cov_0)
    mu_0 = mu_0 + kal_0 @ (y[0] - l_mat_0 @ mu_0)
    cov_0 = cov_0 - kal_0 @ l_mat_0 @ cov_0

    def body_fun(carry: Tuple[jnp.ndarray, jnp.ndarray],
                 i: int) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], Tuple[jnp.ndarray, jnp.ndarray]]:
        mu_tmin1, cov_tmin1 = carry

        t_mat = lgssm_scenario.get_transition_matrix(t[i - 1], t[i])
        t_cov_sqrt = lgssm_scenario.get_transition_covariance_sqrt(t[i - 1], t[i])
        t_cov = t_cov_sqrt @ t_cov_sqrt.T

        # Predict
        mu_t_given_tmin1 = t_mat @ mu_tmin1
        cov_t_given_tmin1 = t_mat @ cov_tmin1 @ t_mat.T + t_cov

        # Update
        l_mat = lgssm_scenario.get_likelihood_matrix(t[i])
        l_cov_sqrt = lgssm_scenario.get_likelihood_covariance_sqrt(t[i])
        l_cov = l_cov_sqrt @ l_cov_sqrt.T

        kal_gain = cov_t_given_tmin1 @ l_mat.T @ jnp.linalg.inv(l_mat @ cov_t_given_tmin1 @ l_mat.T + l_cov)
        mu_t = mu_t_given_tmin1 + kal_gain @ (y[i] - l_mat @ mu_t_given_tmin1)
        cov_t = cov_t_given_tmin1 - kal_gain @ l_mat @ cov_t_given_tmin1

        return (mu_t, cov_t), (mu_t, cov_t)

    _, (mus, covs) = scan(body_fun,
                          (mu_0, cov_0),
                          jnp.arange(1, len(t)))

    return jnp.append(mu_0[jnp.newaxis], mus, 0), jnp.append(cov_0[jnp.newaxis], covs, 0)


def run_kalman_smoother_for_marginals_and_lag_ones(lgssm_scenario: LinearGaussian,
                                                   y: jnp.ndarray,
                                                   t: jnp.ndarray,
                                                   filter_output: Tuple[jnp.ndarray, jnp.ndarray] = None) \
        -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    if filter_output is None:
        filter_output = run_kalman_filter_for_marginals(lgssm_scenario, y, t)

    f_mus, f_covs = filter_output

    def body_fun(carry: Tuple[jnp.ndarray, jnp.ndarray],
                 i: int) -> Tuple[Tuple[jnp.ndarray, jnp.ndarray], Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]:
        mu_tplus1, cov_tplus1 = carry

        t_mat = lgssm_scenario.get_transition_matrix(t[i], t[i + 1])
        t_cov_sqrt = lgssm_scenario.get_transition_covariance_sqrt(t[i], t[i + 1])
        t_cov = t_cov_sqrt @ t_cov_sqrt.T

        f_mu_t = f_mus[i]
        f_cov_t = f_covs[i]
        back_kal_gain = f_cov_t @ t_mat.T @ jnp.linalg.inv(t_mat @ f_cov_t @ t_mat.T + t_cov)
        mu_t = f_mu_t + back_kal_gain @ (mu_tplus1 - t_mat @ f_mu_t)
        cov_t = f_cov_t + back_kal_gain @ (cov_tplus1 - t_mat @ f_cov_t @ t_mat.T - t_cov) @ back_kal_gain.T

        e_xt_xtp1 = f_mu_t @ mu_tplus1.T + back_kal_gain @ (cov_tplus1 + (mu_tplus1 - f_mus[i + 1]) @ mu_tplus1.T)

        return (mu_t, cov_t), (mu_t, cov_t, e_xt_xtp1)

    _, (mus, covs, lag_one_covs) = scan(body_fun,
                                        (f_mus[-1], f_covs[-1]),
                                        jnp.arange(len(t) - 2, -1, -1))

    return jnp.append(mus[::-1], f_mus[-1, jnp.newaxis], 0), \
           jnp.append(covs[::-1], f_covs[-1, jnp.newaxis], 0), \
           lag_one_covs[::-1]


def run_kalman_smoother_for_marginals(lgssm_scenario: LinearGaussian,
                                      y: jnp.ndarray,
                                      t: jnp.ndarray,
                                      filter_output: Tuple[jnp.ndarray, jnp.ndarray] = None) \
        -> Tuple[jnp.ndarray, jnp.ndarray]:
    mus, covs, _ = run_kalman_smoother_for_marginals_and_lag_ones(lgssm_scenario, y, t, filter_output)
    return mus, covs



########################################################################################################################
# Module: abc/importance.py
# Description: Importance (or rejection) sampling for ABC, weight non-zero if simulated summary statistic hits ball.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################


from typing import Tuple

from jax import numpy as jnp, random
from mocat.src.abc.abc import ABCScenario, ABCSampler
from mocat.src.core import cdict, is_implemented


class ImportanceABC(ABCSampler):
    name = 'Importance ABC'

    def importance_proposal(self,
                            abc_scenario: ABCScenario,
                            random_key: jnp.ndarray) -> jnp.ndarray:
        raise NotImplementedError(f'{self.__class__.__name__} importance_proposal not initiated')

    def importance_potential(self,
                             abc_scenario: ABCScenario,
                             x: jnp.ndarray) -> float:
        raise AttributeError(f'{self.__class__.__name__} importance_potential not initiated')

    def log_weight(self,
                   abc_scenario: ABCScenario,
                   state: cdict,
                   extra: cdict):
        return jnp.where(state.distance < extra.parameters.threshold,
                         -state.prior_potential + self.importance_potential(abc_scenario, state.value),
                         -jnp.inf)

    def startup(self,
                abc_scenario: ABCScenario,
                n: int,
                initial_state: cdict = None,
                initial_extra: cdict = None,
                **kwargs) -> Tuple[cdict, cdict]:
        if initial_state is None:
            if is_implemented(abc_scenario.prior_sample):
                initial_extra.random_key, sub_key = random.split(initial_extra.random_key)
                init_vals = abc_scenario.prior_sample(sub_key)
            else:
                init_vals = jnp.zeros(abc_scenario.dim)
            initial_state = cdict(value=init_vals)

        self.max_iter = n - 1

        initial_state, initial_extra = super().startup(abc_scenario, n,
                                                       initial_state, initial_extra, **kwargs)

        initial_state.log_weight = self.log_weight(abc_scenario, initial_state, initial_extra)
        return initial_state, initial_extra

    def proposal(self,
                 abc_scenario: ABCScenario,
                 reject_state: cdict,
                 reject_extra: cdict) -> Tuple[cdict, cdict]:
        proposed_state = reject_state.copy()
        proposed_extra = reject_extra.copy()
        proposed_extra.random_key, subkey1, subkey2, subkey3 = random.split(reject_extra.random_key, 4)
        proposed_state.value = self.importance_proposal(abc_scenario, subkey1)
        proposed_state.prior_potential = abc_scenario.prior_potential(proposed_state.value, subkey2)
        proposed_state.simulated_data = abc_scenario.likelihood_sample(proposed_state.value, subkey3)
        proposed_state.distance = abc_scenario.distance_function(proposed_state.simulated_data)
        proposed_state.log_weight = self.log_weight(abc_scenario, proposed_state, proposed_extra)
        return proposed_state, proposed_extra

    def update(self,
               abc_scenario: ABCScenario,
               state: cdict,
               extra: cdict) -> Tuple[cdict, cdict]:
        proposed_state, proposed_extra = self.proposal(abc_scenario, state, extra)
        proposed_state.log_weight = self.log_weight(abc_scenario, proposed_state, proposed_extra)
        return proposed_state, proposed_extra


class VanillaABC(ImportanceABC):
    name = 'Vanilla ABC'

    def __init__(self,
                 threshold: float = jnp.inf,
                 acceptance_rate: float = None,
                 **kwargs):
        super().__init__(**kwargs)
        self.parameters.threshold = threshold
        if acceptance_rate is not None:
            self.parameters.acceptance_rate = acceptance_rate
            self.clean_chain = self.clean_chain_ar

    def importance_proposal(self,
                            abc_scenario: ABCScenario,
                            random_key: jnp.ndarray) -> jnp.ndarray:
        return abc_scenario.prior_sample(random_key)

    def importance_potential(self,
                             abc_scenario: ABCScenario,
                             x: jnp.ndarray) -> float:
        return abc_scenario.prior_potential(x)

    def log_weight(self,
                   abc_scenario: ABCScenario,
                   state: cdict,
                   extra: cdict):
        return jnp.where(state.distance < extra.parameters.threshold,
                         0.,
                         -jnp.inf)

    def clean_chain_ar(self,
                       abc_scenario: ABCScenario,
                       chain_state: cdict):
        threshold = jnp.quantile(chain_state.distance, self.parameters.acceptance_rate)
        self.parameters.threshold = float(threshold)
        chain_state.log_weight = jnp.where(chain_state.distance < threshold,
                                           0.,
                                           -jnp.inf)
        return chain_state



########################################################################################################################
# Module: mcmc/standard_mcmc.py
# Description: Some common MCMC proposals.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple, Union

import jax.numpy as jnp
from jax import random

from mocat.src.core import cdict, Scenario
from mocat.src.mcmc.sampler import MCMCSampler
from mocat.src import utils
from mocat.src.mcmc.metropolis import Metropolis, mh_acceptance_probability


# Random-walk
# Identity pre-conditioner
class RandomWalk(MCMCSampler):
    name = 'Random Walk'
    correction = Metropolis

    def __init__(self,
                 stepsize: float = None):
        super().__init__()
        self.parameters.stepsize = stepsize
        self.tuning.target = 0.234

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        initial_state, initial_extra = super().startup(scenario, n,
                                                       initial_state, initial_extra, **kwargs)
        initial_extra.random_key, scen_key = random.split(initial_extra.random_key)
        initial_state.potential = scenario.potential(initial_state.value, scen_key)
        return initial_state, initial_extra

    def proposal(self,
                 scenario: Scenario,
                 reject_state: cdict,
                 reject_extra: cdict) -> Tuple[cdict, cdict]:
        proposed_state = reject_state.copy()

        d = scenario.dim
        x = reject_state.value

        stepsize = reject_extra.parameters.stepsize

        reject_extra.random_key, subkey, scen_key = random.split(reject_extra.random_key, 3)

        proposed_state.value = x + jnp.sqrt(stepsize) * random.normal(subkey, (d,))
        proposed_state.potential = scenario.potential(proposed_state.value, scen_key)

        return proposed_state, reject_extra

    def acceptance_probability(self,
                               scenario: Scenario,
                               reject_state: cdict, reject_extra: cdict,
                               proposed_state: cdict, proposed_extra: cdict) -> Union[float, jnp.ndarray]:
        return jnp.minimum(1., jnp.exp(- proposed_state.potential + reject_state.potential))


# Underdamped Langevin
# Identity pre-conditioner/mass matrix
# Recover HMC with friction = jnp.inf
# Recover Overdamped with friction = jnp.inf and leapfrog_steps = 1
class Underdamped(MCMCSampler):
    name = 'Underdamped'
    correction = Metropolis

    def __init__(self,
                 stepsize: float = None,
                 leapfrog_steps: int = 1,
                 friction: float = jnp.inf):
        super().__init__()
        self.parameters.stepsize = stepsize
        self.parameters.leapfrog_steps = leapfrog_steps
        self.parameters.friction = friction
        self.tuning.target = 0.651

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        initial_state, initial_extra = super().startup(scenario, n,
                                                       initial_state, initial_extra, **kwargs)
        initial_extra.random_key, prior_key, lik_key = random.split(initial_extra.random_key, 3)
        initial_state.prior_potential, initial_state.grad_prior_potential\
            = scenario.prior_potential_and_grad(initial_state.value, prior_key)
        initial_state.likelihood_potential, initial_state.grad_likelihood_potential\
            = scenario.likelihood_potential_and_grad(initial_state.value, prior_key)
        initial_state.potential\
            = initial_state.prior_potential + scenario.temperature * initial_state.likelihood_potential
        initial_state.grad_potential\
            = initial_state.grad_prior_potential + scenario.temperature * initial_state.grad_likelihood_potential
        if not hasattr(initial_state, 'momenta') or initial_state.momenta.shape[-1] != scenario.dim:
            initial_state.momenta = jnp.zeros(scenario.dim)
        return initial_state, initial_extra

    def always(self,
               scenario: Scenario,
               reject_state: cdict,
               reject_extra: cdict) -> Tuple[cdict, cdict]:
        d = scenario.dim

        stepsize = reject_extra.parameters.stepsize
        friction = reject_extra.parameters.friction

        reject_state.momenta = - reject_state.momenta

        # Update p - exactly according to solution of OU process
        # Accepted even if leapfrog step is rejected
        reject_extra.random_key, subkey = random.split(reject_extra.random_key)
        reject_state.momenta = reject_state.momenta * jnp.exp(- friction * stepsize) \
                               + jnp.sqrt(1 - jnp.exp(- 2 * friction * stepsize)) * random.normal(subkey, (d,))
        return reject_state, reject_extra

    def proposal(self,
                 scenario: Scenario,
                 reject_state: cdict,
                 reject_extra: cdict) -> Tuple[cdict, cdict]:
        random_keys = random.split(reject_extra.random_key, self.parameters.leapfrog_steps + 1)

        reject_extra.random_key = random_keys[0]

        all_leapfrog_state = utils.leapfrog(scenario.prior_potential_and_grad,
                                            scenario.likelihood_potential_and_grad,
                                            reject_state,
                                            reject_extra.parameters.stepsize,
                                            random_keys[1:],
                                            scenario.temperature)
        proposed_state = all_leapfrog_state[-1]

        proposed_state.momenta = - proposed_state.momenta

        return proposed_state, reject_extra

    def acceptance_probability(self,
                               scenario: Scenario,
                               reject_state: cdict, reject_extra: cdict,
                               proposed_state: cdict, proposed_extra: cdict) -> Union[float, jnp.ndarray]:
        pre_min_alpha = jnp.exp(- proposed_state.potential
                                + reject_state.potential
                                - utils.gaussian_potential(proposed_state.momenta)
                                + utils.gaussian_potential(reject_state.momenta))
        return jnp.minimum(1., pre_min_alpha)





from mocat import utils
from mocat import kernels
from mocat import scenarios
from mocat import twodim
from mocat import transport
from mocat import ssm
from mocat import abc

from mocat.version import __version__

from mocat.src.core import Scenario
from mocat.src.core import cdict
from mocat.src.core import static_cdict
from mocat.src.core import save_cdict
from mocat.src.core import load_cdict

from mocat.src.sample import Sampler
from mocat.src.sample import run

from mocat.src.mcmc.sampler import MCMCSampler
from mocat.src.mcmc.sampler import Correction
from mocat.src.mcmc.sampler import Uncorrected

from mocat.src.mcmc.metropolis import Metropolis
from mocat.src.mcmc.metropolis import RMMetropolis

from mocat.src.mcmc.standard_mcmc import RandomWalk
from mocat.src.mcmc.standard_mcmc import Underdamped
from mocat.src.mcmc.qn_underdamped import QNUnderdamped

from mocat.src.transport.sampler import TransportSampler

from mocat.src.transport.smc import SMCSampler
from mocat.src.transport.smc import TemperedSMCSampler
from mocat.src.transport.smc import MetropolisedSMCSampler
from mocat.src.transport.smc import RMMetropolisedSMCSampler

from mocat.src.transport.svgd import SVGD

from mocat.src.transport.teki import TemperedEKI
from mocat.src.transport.teki import AdaptiveTemperedEKI


from mocat.src.metrics import autocorrelation
from mocat.src.metrics import integrated_autocorrelation_time
from mocat.src.metrics import ess_autocorrelation
from mocat.src.metrics import log_ess_log_weight
from mocat.src.metrics import ess_log_weight
from mocat.src.metrics import squared_jumping_distance
from mocat.src.metrics import ksd
from mocat.src.metrics import metric_plot
from mocat.src.metrics import autocorrelation_plot
from mocat.src.metrics import trace_plot
from mocat.src.metrics import plot_2d_samples
from mocat.src.metrics import hist_1d_samples


try:
  del src
except NameError:
  pass




########################################################################################################################
# Module: ssm/online_smoothing.py
# Description: Online (fixed-lag) particle smoothing for state-space models.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple
from functools import partial

from jax import numpy as jnp, random, vmap, jit
from jax.lax import while_loop, cond, map

from mocat.src.core import cdict
from mocat.src.ssm.ssm import StateSpaceModel
from mocat.src.ssm.filtering import ParticleFilter, resample_particles, propagate_particle_filter
from mocat.src.ssm.backward import backward_simulation
from mocat.src.metrics import ess_log_weight


def full_stitch_single(ssm_scenario: StateSpaceModel,
                       x0_single: jnp.ndarray,
                       t: float,
                       x1_all: jnp.ndarray,
                       tplus1: float,
                       x1_log_weight: jnp.ndarray,
                       random_key: jnp.ndarray) -> jnp.ndarray:
    log_weight = x1_log_weight - vmap(ssm_scenario.transition_potential, (None, None, 0, None))(x0_single, t,
                                                                                                x1_all, tplus1)
    return random.categorical(random_key, log_weight)


def full_stitch(ssm_scenario: StateSpaceModel,
                x0_all: jnp.ndarray,
                t: float,
                x1_all: jnp.ndarray,
                tplus1: float,
                x1_log_weight: jnp.ndarray,
                random_key: jnp.ndarray) -> jnp.ndarray:
    return vmap(full_stitch_single, (None, 0, None, None, None, None, 0)) \
        (ssm_scenario, x0_all, t, x1_all, tplus1, x1_log_weight, random.split(random_key, len(x0_all)))


def full_stitch_single_cond(not_yet_accepted: bool,
                            x1_ind_false: jnp.ndarray,
                            ssm_scenario: StateSpaceModel,
                            x0_single: jnp.ndarray,
                            t: float,
                            x1_all: jnp.ndarray,
                            tplus1: float,
                            x1_log_weight: jnp.ndarray,
                            random_key: jnp.ndarray) -> jnp.ndarray:
    return cond(not_yet_accepted,
                lambda _: full_stitch_single(ssm_scenario, x0_single, t, x1_all, tplus1, x1_log_weight, random_key),
                lambda _: x1_ind_false,
                None)


def rejection_stitch_proposal_single(ssm_scenario: StateSpaceModel,
                                     x0_single: jnp.ndarray,
                                     t: float,
                                     x1_all: jnp.ndarray,
                                     tplus1: float,
                                     x1_log_weight: jnp.ndarray,
                                     bound: float,
                                     random_key: jnp.ndarray) \
        -> Tuple[jnp.ndarray, float, bool, jnp.ndarray]:
    random_key, choice_key, uniform_key = random.split(random_key, 3)
    x1_single_ind = random.categorical(choice_key, x1_log_weight)
    conditional_dens = jnp.exp(-ssm_scenario.transition_potential(x0_single, t, x1_all[x1_single_ind], tplus1))
    return x1_single_ind, conditional_dens, random.uniform(uniform_key) > conditional_dens / bound, random_key


def rejection_stitch_proposal_single_cond(not_yet_accepted: bool,
                                          x1_ind_false: jnp.ndarray,
                                          ssm_scenario: StateSpaceModel,
                                          x0_single: jnp.ndarray,
                                          t: float,
                                          x1_all: jnp.ndarray,
                                          tplus1: float,
                                          x1_log_weight: jnp.ndarray,
                                          bound: float,
                                          random_key: jnp.ndarray) -> Tuple[jnp.ndarray, float, bool, jnp.ndarray]:
    return cond(not_yet_accepted,
                lambda _: rejection_stitch_proposal_single(ssm_scenario,
                                                           x0_single, t,
                                                           x1_all, tplus1,
                                                           x1_log_weight, bound, random_key),
                lambda _: (x1_ind_false, 0., False, random_key),
                None)


def rejection_stitch_proposal_all(ssm_scenario: StateSpaceModel,
                                  x0_all: jnp.ndarray,
                                  t: float,
                                  x1_all: jnp.ndarray,
                                  tplus1: float,
                                  x1_log_weight: jnp.ndarray,
                                  bound_inflation: float,
                                  not_yet_accepted_arr: jnp.ndarray,
                                  x1_all_sampled_inds: jnp.ndarray,
                                  bound: float,
                                  random_keys: jnp.ndarray,
                                  rejection_iter: int,
                                  num_transition_evals: int) \
        -> Tuple[jnp.ndarray, jnp.ndarray, float, jnp.ndarray, int, int]:
    n = len(x1_all)
    mapped_tup = map(lambda i: rejection_stitch_proposal_single_cond(not_yet_accepted_arr[i],
                                                                     x1_all_sampled_inds[i],
                                                                     ssm_scenario,
                                                                     x0_all[i],
                                                                     t,
                                                                     x1_all,
                                                                     tplus1,
                                                                     x1_log_weight,
                                                                     bound,
                                                                     random_keys[i]), jnp.arange(n))
    x1_all_sampled_inds, dens_evals, not_yet_accepted_arr_new, random_keys = mapped_tup

    # Check if we need to start again
    max_dens = jnp.max(dens_evals)
    reset_bound = max_dens > bound
    bound = jnp.where(reset_bound, max_dens * bound_inflation, bound)
    not_yet_accepted_arr_new = jnp.where(reset_bound, jnp.ones(n, dtype='bool'), not_yet_accepted_arr_new)
    return not_yet_accepted_arr_new, x1_all_sampled_inds, bound, random_keys, rejection_iter + 1, \
           num_transition_evals + not_yet_accepted_arr.sum()


def rejection_stitching(ssm_scenario: StateSpaceModel,
                        x0_all: jnp.ndarray,
                        t: float,
                        x1_all: jnp.ndarray,
                        tplus1: float,
                        x1_log_weight: jnp.ndarray,
                        random_key: jnp.ndarray,
                        maximum_rejections: int,
                        init_bound_param: float,
                        bound_inflation: float) -> Tuple[jnp.ndarray, int]:
    rejection_initial_keys = random.split(random_key, 3)
    n = len(x1_all)

    # Prerun to initiate bound
    x1_initial_inds = random.categorical(rejection_initial_keys[0], x1_log_weight, shape=(n,))
    initial_cond_dens = jnp.exp(-vmap(ssm_scenario.transition_potential,
                                      (0, None, 0, None))(x0_all, t, x1_all[x1_initial_inds], tplus1))
    max_cond_dens = jnp.max(initial_cond_dens)
    initial_bound = jnp.where(max_cond_dens > init_bound_param, max_cond_dens * bound_inflation, init_bound_param)
    initial_not_yet_accepted_arr = random.uniform(rejection_initial_keys[1], (n,)) > initial_cond_dens / initial_bound

    out_tup = while_loop(lambda tup: jnp.logical_and(tup[0].sum() > 0, tup[-2] < maximum_rejections),
                         lambda tup: rejection_stitch_proposal_all(ssm_scenario, x0_all, t, x1_all, tplus1,
                                                                   x1_log_weight,
                                                                   bound_inflation, *tup),
                         (initial_not_yet_accepted_arr,
                          x1_initial_inds,
                          initial_bound,
                          random.split(rejection_initial_keys[2], n),
                          1,
                          n))
    not_yet_accepted_arr, x1_final_inds, final_bound, random_keys, rej_attempted, num_transition_evals = out_tup

    x1_final_inds = map(lambda i: full_stitch_single_cond(not_yet_accepted_arr[i],
                                                          x1_final_inds[i],
                                                          ssm_scenario,
                                                          x0_all[i],
                                                          t,
                                                          x1_all,
                                                          tplus1,
                                                          x1_log_weight,
                                                          random_keys[i]), jnp.arange(n))

    num_transition_evals = num_transition_evals + len(x1_all) * not_yet_accepted_arr.sum()

    return x1_final_inds, num_transition_evals


@partial(jit, static_argnums=(0,))
def fixed_lag_stitching(ssm_scenario: StateSpaceModel,
                        early_block: jnp.ndarray,
                        t: float,
                        recent_block: jnp.ndarray,
                        recent_block_log_weight: jnp.ndarray,
                        tplus1: float,
                        random_key: jnp.ndarray,
                        maximum_rejections: int,
                        init_bound_param: float,
                        bound_inflation: float) -> Tuple[jnp.ndarray, int]:
    x0_fixed_all = early_block[-1]

    x0_vary_all = recent_block[0]
    x1_vary_all = recent_block[1]

    non_interacting_log_weight = recent_block_log_weight \
                                 + vmap(ssm_scenario.transition_potential, (0, None, 0, None))(x0_vary_all, t,
                                                                                               x1_vary_all, tplus1)

    recent_stitched_inds, num_transition_evals \
        = cond(maximum_rejections > 0,
               lambda tup: rejection_stitching(ssm_scenario, *tup,
                                               maximum_rejections=maximum_rejections,
                                               init_bound_param=init_bound_param,
                                               bound_inflation=bound_inflation),
               lambda tup: (
                   full_stitch(ssm_scenario, *tup), len(x0_fixed_all) ** 2),
               (x0_fixed_all, t, x1_vary_all, tplus1,
                non_interacting_log_weight, random_key))

    return jnp.append(early_block, recent_block[1:, recent_stitched_inds], axis=0), num_transition_evals


@partial(jit, static_argnums=(0, 1, 6, 7))
def propagate_particle_smoother_pf(ssm_scenario: StateSpaceModel,
                                   particle_filter: ParticleFilter,
                                   particles: cdict,
                                   y_new: jnp.ndarray,
                                   t_new: float,
                                   random_key: jnp.ndarray,
                                   lag: int,
                                   maximum_rejections: int,
                                   init_bound_param: float,
                                   bound_inflation: float) -> cdict:
    if not hasattr(particles, 'num_transition_evals'):
        particles.num_transition_evals = jnp.array(0)

    n = particles.value.shape[1]

    # Check particles are unweighted
    out_particles = cond(ess_log_weight(jnp.atleast_2d(particles.log_weight)[-1]) < (n - 1e-3),
                         lambda p: resample_particles(p, random_key, True),
                         lambda p: p.copy(),
                         particles)
    out_particles.log_weight = jnp.zeros(n)

    x_previous = out_particles.value[-1]
    t_previous = out_particles.t[-1]

    split_keys = random.split(random_key, len(x_previous))

    x_new, out_particles.log_weight = particle_filter.propose_and_intermediate_weight_vectorised(ssm_scenario,
                                                                                                 x_previous, t_previous,
                                                                                                 y_new, t_new,
                                                                                                 split_keys)

    out_particles.value = jnp.append(out_particles.value, x_new[jnp.newaxis], axis=0)
    out_particles.y = jnp.append(out_particles.y, y_new[jnp.newaxis], axis=0)
    out_particles.t = jnp.append(out_particles.t, t_new)
    out_particles.ess = ess_log_weight(out_particles.log_weight)

    len_t = len(out_particles.t)
    stitch_ind_min_1 = len_t - lag - 1
    stitch_ind = len_t - lag

    # out_particles.value = cond(stitch_ind_min_1 >= 0,
    #                            lambda vals: fixed_lag_stitching(ssm_scenario,
    #                                                             vals[:(stitch_ind_min_1 + 1)],
    #                                                             out_particles.t[stitch_ind_min_1],
    #                                                             vals[stitch_ind_min_1:],
    #                                                             out_particles.log_weight,
    #                                                             out_particles.t[stitch_ind],
    #                                                             random_key,
    #                                                             maximum_rejections,
    #                                                             init_bound_param,
    #                                                             bound_inflation),
    #                            lambda vals: vals,
    #                            out_particles.value)

    num_transition_evals = 0
    if stitch_ind_min_1 >= 0:
        out_particles.value, num_transition_evals = fixed_lag_stitching(ssm_scenario,
                                                                        out_particles.value[:(stitch_ind_min_1 + 1)],
                                                                        out_particles.t[stitch_ind_min_1],
                                                                        out_particles.value[stitch_ind_min_1:],
                                                                        out_particles.log_weight,
                                                                        out_particles.t[stitch_ind],
                                                                        random_key,
                                                                        maximum_rejections,
                                                                        init_bound_param,
                                                                        bound_inflation)
    out_particles.num_transition_evals = jnp.append(out_particles.num_transition_evals, num_transition_evals)
    out_particles.log_weight = jnp.where(stitch_ind_min_1 >= 0, jnp.zeros(n), out_particles.log_weight)
    return out_particles


@partial(jit, static_argnums=(0, 1, 6, 8))
def propagate_particle_smoother_bs(ssm_scenario: StateSpaceModel,
                                   particle_filter: ParticleFilter,
                                   particles: cdict,
                                   y_new: jnp.ndarray,
                                   t_new: float,
                                   random_key: jnp.ndarray,
                                   lag: int,
                                   ess_threshold: float,
                                   maximum_rejections: int,
                                   init_bound_param: float,
                                   bound_inflation: float) -> cdict:
    n = particles.value.shape[1]

    if not hasattr(particles, 'num_transition_evals'):
        particles.num_transition_evals = jnp.array(0)

    if not hasattr(particles, 'marginal_filter'):
        particles.marginal_filter = cdict(value=particles.value,
                                          log_weight=particles.log_weight,
                                          y=particles.y,
                                          t=particles.t,
                                          ess=particles.ess)

    split_keys = random.split(random_key, 4)

    out_particles = particles

    # Propagate marginal filter particles
    out_particles.marginal_filter = propagate_particle_filter(ssm_scenario, particle_filter, particles.marginal_filter,
                                                              y_new, t_new, split_keys[1], ess_threshold, False)
    out_particles.y = jnp.append(out_particles.y, y_new[jnp.newaxis], axis=0)
    out_particles.t = jnp.append(out_particles.t, t_new)
    out_particles.log_weight = jnp.zeros(n)
    out_particles.ess = out_particles.marginal_filter.ess[-1]

    len_t = len(out_particles.t)
    stitch_ind_min_1 = len_t - lag - 1
    stitch_ind = len_t - lag

    def back_sim_only(marginal_filter):
        backward_sim = backward_simulation(ssm_scenario,
                                           marginal_filter,
                                           split_keys[2],
                                           n,
                                           maximum_rejections,
                                           init_bound_param,
                                           bound_inflation)
        return backward_sim.value, backward_sim.num_transition_evals.sum()

    def back_sim_and_stitch(marginal_filter):
        backward_sim = backward_simulation(ssm_scenario,
                                           marginal_filter[stitch_ind_min_1:],
                                           split_keys[2],
                                           n,
                                           maximum_rejections,
                                           init_bound_param,
                                           bound_inflation)

        vals, stitch_nte = fixed_lag_stitching(ssm_scenario,
                                               out_particles.value[:(stitch_ind_min_1 + 1)],
                                               out_particles.t[stitch_ind_min_1],
                                               backward_sim.value,
                                               jnp.zeros(n),
                                               out_particles.t[stitch_ind],
                                               random_key,
                                               maximum_rejections,
                                               init_bound_param,
                                               bound_inflation)
        return vals, stitch_nte + backward_sim.num_transition_evals.sum()

    if stitch_ind_min_1 >= 0:
        out_particles.value, num_transition_evals = back_sim_and_stitch(out_particles.marginal_filter)
    else:
        out_particles.value, num_transition_evals = back_sim_only(out_particles.marginal_filter)

    out_particles.num_transition_evals = jnp.append(out_particles.num_transition_evals, num_transition_evals)

    # out_particles.value = cond(stitch_ind_min_1 >= 0,
    #                            back_sim_and_stitch,
    #                            back_sim_only,
    #                            out_particles.marginal_filter)
    return out_particles


def propagate_particle_smoother(ssm_scenario: StateSpaceModel,
                                particle_filter: ParticleFilter,
                                particles: cdict,
                                y_new: jnp.ndarray,
                                t_new: float,
                                random_key: jnp.ndarray,
                                lag: int,
                                backward_sim: bool = True,
                                ess_threshold: float = 0.5,
                                maximum_rejections: int = 0,
                                init_bound_param: float = 0.,
                                bound_inflation: float = 1.01) -> cdict:
    if backward_sim:
        return propagate_particle_smoother_bs(ssm_scenario, particle_filter, particles, y_new, t_new, random_key, lag,
                                              ess_threshold, maximum_rejections, init_bound_param, bound_inflation)
    else:
        return propagate_particle_smoother_pf(ssm_scenario, particle_filter, particles, y_new, t_new, random_key, lag,
                                              maximum_rejections, init_bound_param, bound_inflation)




from mocat.src.ssm.scenarios import scenarios

from mocat.src.ssm.ssm import StateSpaceModel

from mocat.src.ssm.filtering import ParticleFilter
from mocat.src.ssm.filtering import BootstrapFilter
from mocat.src.ssm.filtering import initiate_particles
from mocat.src.ssm.filtering import resample_particles
from mocat.src.ssm.filtering import propagate_particle_filter
from mocat.src.ssm.filtering import run_particle_filter_for_marginals

from mocat.src.ssm.backward import backward_simulation
from mocat.src.ssm.backward import forward_filtering_backward_simulation

from mocat.src.ssm.online_smoothing import propagate_particle_smoother
from mocat.src.ssm.online_smoothing import propagate_particle_smoother_pf
from mocat.src.ssm.online_smoothing import propagate_particle_smoother_bs

from mocat.src.ssm.linear_gaussian.linear_gaussian import LinearGaussian
from mocat.src.ssm.linear_gaussian.linear_gaussian import TimeHomogenousLinearGaussian
from mocat.src.ssm.linear_gaussian.kalman import run_kalman_filter_for_marginals
from mocat.src.ssm.linear_gaussian.kalman import run_kalman_smoother_for_marginals
from mocat.src.ssm.linear_gaussian.kalman import run_kalman_smoother_for_marginals_and_lag_ones


from mocat.src.ssm.nonlinear_gaussian import NonLinearGaussian
from mocat.src.ssm.nonlinear_gaussian import OptimalNonLinearGaussianParticleFilter
from mocat.src.ssm.nonlinear_gaussian import EnsembleKalmanFilter





########################################################################################################################
# Module: abc/mcmc.py
# Description: Markovian ABC, accept if simulated data hits ball.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################
from typing import Union, Tuple

from jax import numpy as jnp, random
from mocat.src.abc.abc import ABCScenario, ABCSampler
from mocat.src.core import cdict
from mocat.src.mcmc.metropolis import Metropolis, RMMetropolis
from mocat.src.mcmc.sampler import MCMCSampler, Correction


class ABCMCMCSampler(ABCSampler, MCMCSampler):
    name = "ABC MCMC"
    correction: Correction = Metropolis
    tuning: cdict = None

    def __init__(self,
                 threshold: float = jnp.inf,
                 **kwargs):
        super().__init__(threshold=threshold, **kwargs)

    def always(self,
               abc_scenario: ABCScenario,
               reject_state: cdict,
               reject_extra: cdict) -> Tuple[cdict, cdict]:
        reject_extra.random_key, _ = random.split(reject_extra.random_key)
        return reject_state, reject_extra

    def proposal(self,
                 abc_scenario: ABCScenario,
                 reject_state: cdict,
                 reject_extra: cdict) -> Tuple[cdict, cdict]:
        raise NotImplementedError(f'{self.__class__.__name__} markov proposal not initiated')


class RandomWalkABC(ABCMCMCSampler):
    name = 'Random Walk ABC'
    correction = Metropolis

    def __init__(self,
                 threshold: float = None,
                 stepsize: float = None):
        super().__init__()
        self.parameters.threshold = threshold
        self.parameters.stepsize = stepsize

        self.tuning = cdict(parameter='threshold',
                            target=0.1,
                            metric='alpha',
                            monotonicity='increasing')

    def acceptance_probability(self,
                               abc_scenario: ABCScenario,
                               reject_state: cdict, reject_extra: cdict,
                               proposed_state: cdict, proposed_extra: cdict) -> Union[float, jnp.ndarray]:
        return jnp.minimum(1., jnp.exp(-proposed_state.prior_potential
                                       + reject_state.prior_potential)
                           * (proposed_state.distance < reject_extra.parameters.threshold))

    def proposal(self,
                 abc_scenario: ABCScenario,
                 reject_state: cdict,
                 reject_extra: cdict) -> Tuple[cdict, cdict]:
        proposed_state = reject_state.copy()
        proposed_extra = reject_extra.copy()
        stepsize = reject_extra.parameters.stepsize
        proposed_extra.random_key, subkey1, subkey2, subkey3 = random.split(reject_extra.random_key, 4)
        proposed_state.value = reject_state.value + jnp.sqrt(stepsize) * random.normal(subkey1, (abc_scenario.dim,))
        proposed_state.prior_potential = abc_scenario.prior_potential(proposed_state.value, subkey2)
        proposed_state.simulated_data = abc_scenario.likelihood_sample(proposed_state.value, subkey3)
        proposed_state.distance = abc_scenario.distance_function(proposed_state.simulated_data)
        return proposed_state, proposed_extra


class RMMetropolisDiagStepsize(RMMetropolis):

    def startup(self,
                abc_scenario: ABCScenario,
                sampler: MCMCSampler,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        initial_state, initial_extra = super().startup(abc_scenario, sampler, n,
                                                       initial_state, initial_extra, **kwargs)
        if sampler.parameters.stepsize is None:
            initial_extra.parameters.stepsize = 1.0
        if sampler.parameters.threshold is None:
            initial_extra.parameters.threshold = 50.
            initial_state.threshold = initial_extra.parameters.threshold
        initial_extra.parameters.stepsize = jnp.ones(abc_scenario.dim) * initial_extra.parameters.stepsize
        initial_state.stepsize = initial_extra.parameters.stepsize
        initial_extra.post_mean = initial_state.value
        initial_extra.diag_post_cov = initial_extra.parameters.stepsize * abc_scenario.dim / 2.38 ** 2
        return initial_state, initial_extra

    def adapt(self,
              state: cdict,
              extra: cdict) -> Tuple[cdict, cdict]:
        # Adapt threshold
        adapted_state, adapted_extra = super().adapt(state, extra)

        # Adapt stepsize to be 2.28^2 / d * diagonal covariance of samples
        rm_stepsize = self.rm_stepsize_scale * extra.iter ** -self.rm_stepsize_neg_exponent
        rm_stepsize = jnp.minimum(rm_stepsize, 1 - 1e-3)

        d = adapted_extra.diag_post_cov.size

        adapted_extra.diag_post_cov = (1 - rm_stepsize) * adapted_extra.diag_post_cov \
                                      + rm_stepsize * (adapted_state.value - adapted_extra.post_mean) ** 2

        adapted_extra.post_mean = (1 - rm_stepsize) * adapted_extra.post_mean + rm_stepsize * adapted_state.value

        adapted_extra.parameters.stepsize = adapted_extra.diag_post_cov / d * 2.38 ** 2
        adapted_state.stepsize = adapted_extra.parameters.stepsize
        return adapted_state, adapted_extra



########################################################################################################################
# Module: transport/svgd.py
# Description: Stein Variational Gradient Descent.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Tuple, Union, Callable

from jax import numpy as jnp, vmap, random
from jax.example_libraries.optimizers import adagrad

from mocat.src.core import Scenario, cdict, impl_checkable, is_implemented
from mocat.src.transport.sampler import TransportSampler
from mocat.src.kernels import Kernel, Gaussian, mean_bandwidth_update


def kernelised_grad_matrix(vals: jnp.ndarray,
                           grads: jnp.ndarray,
                           kernel: Kernel,
                           kernel_params: cdict,
                           batch_inds_mat: jnp.ndarray) -> jnp.ndarray:
    n = vals.shape[0]

    def phi_hat_func(x_i):
        return vmap(lambda x_j: -kernel._call(vals[x_j], vals[x_i], **kernel_params)
                                * grads[x_j]
                                + kernel._grad_x(vals[x_j], vals[x_i], **kernel_params)
                    )(batch_inds_mat[x_i]).mean(axis=0)

    phi_hat = vmap(phi_hat_func)(jnp.arange(n))
    return phi_hat


def adapt_bandwidth_mean(ensemble_state: cdict,
                         extra: cdict) -> Tuple[cdict, cdict]:
    extra.parameters.kernel_params.bandwidth = mean_bandwidth_update(ensemble_state.value)
    ensemble_state.kernel_params = extra.parameters.kernel_params
    return ensemble_state, extra


class SVGD(TransportSampler):
    name = 'SVGD'

    opt_init: Callable
    opt_update: Callable
    get_params: Callable
    get_batch_inds: Callable

    def __init__(self,
                 stepsize: Union[float, Callable],
                 max_iter: int = 1000,
                 kernel: Kernel = None,
                 kernel_params: cdict = None,
                 ensemble_batchsize: int = None,
                 optimiser: Callable = adagrad,
                 **optim_params):
        super().__init__(max_iter=max_iter)
        if kernel is None:
            kernel = Gaussian()
            if not is_implemented(self.adapt):
                self.adapt = adapt_bandwidth_mean

        if kernel_params is None:
            kernel_params = kernel.parameters

        self.parameters.stepsize = stepsize

        self.kernel = kernel
        self.parameters.kernel_params = kernel_params

        self.parameters.ensemble_batchsize = ensemble_batchsize

        self.optimiser = optimiser
        self.parameters.optim_params = optim_params

    def startup(self,
                scenario: Scenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:
        initial_state, initial_extra = super().startup(scenario, n, initial_state, initial_extra, **kwargs)

        if self.parameters.ensemble_batchsize is None:
            self.parameters.ensemble_batchsize = n
            initial_extra.parameters.ensemble_batchsize = n

        if self.parameters.ensemble_batchsize == n:
            self.get_batch_inds = lambda _: jnp.repeat(jnp.arange(n)[None], n, axis=0)
        else:
            self.get_batch_inds = lambda rk: random.choice(rk, n, shape=(n, self.parameters.ensemble_batchsize,))

        del initial_extra.parameters.stepsize

        random_keys = random.split(initial_extra.random_key, n + 1)
        initial_extra.random_key = random_keys[-1]

        initial_state.potential, initial_state.grad_potential = vmap(scenario.potential_and_grad)(initial_state.value,
                                                                                                  random_keys[:n])

        initial_state, initial_extra = self.adapt(initial_state, initial_extra)

        self.opt_init, self.opt_update, self.get_params = self.optimiser(step_size=self.parameters.stepsize,
                                                                         **initial_extra.parameters.optim_params)
        initial_extra.opt_state = self.opt_init(initial_state.value)
        return initial_state, initial_extra

    @impl_checkable
    def adapt(self,
              ensemble_state: cdict,
              extra: cdict) -> Tuple[cdict, cdict]:
        return ensemble_state, extra

    def kernelised_grad_matrix(self,
                               vals: jnp.ndarray,
                               grads: jnp.ndarray,
                               kernel_params: cdict,
                               batch_inds_mat: jnp.ndarray) -> jnp.ndarray:
        return kernelised_grad_matrix(vals, grads, self.kernel, kernel_params, batch_inds_mat)

    def update(self,
               scenario: Scenario,
               ensemble_state: cdict,
               extra: cdict) -> Tuple[cdict, cdict]:
        n = ensemble_state.value.shape[0]
        extra.iter = extra.iter + 1

        random_keys = random.split(extra.random_key, n + 2)
        batch_inds = self.get_batch_inds(random_keys[-1])
        extra.random_key = random_keys[-2]

        phi_hat = self.kernelised_grad_matrix(ensemble_state.value,
                                              ensemble_state.grad_potential,
                                              extra.parameters.kernel_params,
                                              batch_inds)

        extra.opt_state = self.opt_update(extra.iter, -phi_hat, extra.opt_state)
        ensemble_state.value = self.get_params(extra.opt_state)

        ensemble_state.potential, ensemble_state.grad_potential \
            = vmap(scenario.potential_and_grad)(ensemble_state.value, random_keys[:n])

        ensemble_state, extra = self.adapt(ensemble_state, extra)

        return ensemble_state, extra



########################################################################################################################
# Module: abc/smc.py
# Description: SMC sampler for ABC problems.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union, Tuple, Type
from inspect import isclass

from jax import numpy as jnp, random, vmap
from jax.lax import scan, cond

from mocat.src.abc.abc import ABCScenario, ABCSampler
from mocat.src.abc.mcmc import ABCMCMCSampler, RandomWalkABC
from mocat.src.core import cdict, is_implemented
from mocat.src.transport.smc import SMCSampler
from mocat.src.mcmc.sampler import Correction, check_correction
from mocat.src.metrics import ess_log_weight


class ABCSMCSampler(ABCSampler, SMCSampler):
    name = "ABC SMC"

    def __init__(self,
                 threshold_schedule: Union[None, jnp.ndarray] = None,
                 max_iter: int = int(1e4),
                 **kwargs):
        self.max_iter = max_iter
        self.threshold_schedule = threshold_schedule
        super().__init__(**kwargs)

    def __setattr__(self, key, value):
        if key == 'threshold_schedule':
            if value is None:
                if hasattr(self, 'threshold_schedule') and self.max_iter == len(self.threshold_schedule):
                    self.max_iter = int(1e4)
                self.next_threshold = self.next_threshold_adaptive
            else:
                self.next_threshold = lambda state, extra: self.threshold_schedule[extra.iter]
                self.max_iter = len(value)
        super().__setattr__(key, value)

    def startup(self,
                abc_scenario: ABCScenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:

        initial_state, initial_extra = SMCSampler.startup(self, abc_scenario, n,
                                                          initial_state, initial_extra, **kwargs)

        n = len(initial_state.value)
        if not hasattr(initial_state, 'prior_potential') and is_implemented(abc_scenario.prior_potential):
            random_keys = random.split(initial_extra.random_key, n + 1)
            initial_extra.random_key = random_keys[-1]
            initial_state.prior_potential = vmap(abc_scenario.prior_potential)(initial_state.value,
                                                                               random_keys[:n])

        if not hasattr(initial_state, 'simulated_data'):
            random_keys = random.split(initial_extra.random_key, n + 1)
            initial_extra.random_key = random_keys[-1]
            initial_state.simulated_data = vmap(abc_scenario.likelihood_sample)(initial_state.value,
                                                                                random_keys[:n])

        if not hasattr(initial_state, 'distance'):
            initial_state.distance = vmap(abc_scenario.distance_function)(initial_state.simulated_data)

        if not hasattr(initial_state, 'threshold'):
            if self.threshold_schedule is None:
                initial_state.threshold = jnp.zeros(n) + jnp.inf
            else:
                initial_state.threshold = jnp.zeros(n) + self.threshold_schedule[0]

        if not hasattr(initial_state, 'ess'):
            initial_state.ess = jnp.zeros(n) + n

        return initial_state, initial_extra

    def next_threshold_adaptive(self,
                                state: cdict,
                                extra: cdict) -> float:
        raise AttributeError(f'{self.name} next_threshold_adaptive not initiated')

    def clean_chain(self,
                    abc_scenario: ABCScenario,
                    chain_ensemble_state: cdict) -> cdict:
        chain_ensemble_state.threshold = chain_ensemble_state.threshold[:, 0]
        chain_ensemble_state.ess = chain_ensemble_state.ess[:, 0]
        return chain_ensemble_state


def adapt_stepsize_scaled_diag_cov(ensemble_state: cdict,
                                   extra: cdict) -> Tuple[cdict, cdict]:
    n, d = ensemble_state.value.shape
    extra.parameters.stepsize = vmap(jnp.cov, (1,))(ensemble_state.value) / d * 2.38 ** 2
    return ensemble_state, extra


class MetropolisedABCSMCSampler(ABCSMCSampler):

    def __init__(self,
                 mcmc_sampler: Union[ABCMCMCSampler, Type[ABCMCMCSampler]] = None,
                 mcmc_correction: Union[Correction, Type[Correction], str] = 'sampler_default',
                 mcmc_steps: int = 1,
                 threshold_schedule: Union[None, jnp.ndarray] = None,
                 max_iter: int = int(1e4),
                 ess_threshold_retain: float = 0.9,
                 ess_threshold_resample: float = 0.5,
                 termination_alpha: float = 0.01,
                 **kwargs):
        super().__init__(max_iter=max_iter, threshold_schedule=threshold_schedule, **kwargs)
        if mcmc_sampler is None:
            mcmc_sampler = RandomWalkABC()
            self.adapt_mcmc_params = lambda ps, pe, ns, ne: adapt_stepsize_scaled_diag_cov(ns, ne)

        if isclass(mcmc_sampler):
            mcmc_sampler = mcmc_sampler()
        self.mcmc_sampler = mcmc_sampler
        if mcmc_correction != 'sampler_default':
            self.mcmc_sampler.correction = mcmc_correction
        self.parameters.mcmc_steps = mcmc_steps
        self.parameters.ess_threshold_retain = ess_threshold_retain
        self.parameters.ess_threshold_resample = ess_threshold_resample
        self.parameters.termination_alpha = termination_alpha

    def startup(self,
                abc_scenario: ABCScenario,
                n: int,
                initial_state: cdict,
                initial_extra: cdict,
                **kwargs) -> Tuple[cdict, cdict]:

        self.mcmc_sampler.correction = check_correction(self.mcmc_sampler.correction)

        initial_state, initial_extra = super().startup(abc_scenario, n, initial_state, initial_extra,
                                                       **kwargs)
        initial_extra.parameters.threshold = initial_state.threshold[0]

        initial_state, initial_extra = vmap(
            lambda state: self.mcmc_sampler.startup(abc_scenario,
                                                    initial_extra.parameters.mcmc_steps,
                                                    state,
                                                    initial_extra))(initial_state)
        initial_extra = initial_extra[0]

        initial_state, initial_extra = self.adapt(initial_state, initial_extra,
                                                  initial_state, initial_extra)
        return initial_state, initial_extra

    def resample_criterion(self,
                           ensemble_state: cdict,
                           extra: cdict) -> bool:
        return ensemble_state.ess[0] < (extra.parameters.ess_threshold_resample * ensemble_state.ess.size)

    def termination_criterion(self,
                              ensemble_state: cdict,
                              extra: cdict) -> bool:
        return jnp.logical_or(extra.alpha_mean <= extra.parameters.termination_alpha,
                              extra.iter >= self.max_iter)

    def next_threshold_adaptive(self,
                                state: cdict,
                                extra: cdict) -> float:
        return jnp.quantile(state.distance, extra.parameters.ess_threshold_retain * state.ess[0] / state.ess.size)

    def log_weight(self,
                   previous_ensemble_state: cdict,
                   previous_extra: cdict,
                   new_ensemble_state: cdict,
                   new_extra: cdict) -> jnp.ndarray:
        return jnp.where(new_ensemble_state.distance > new_extra.parameters.threshold, -jnp.inf, 0.)

    def clean_mcmc_chain(self,
                         chain_state: cdict,
                         chain_extra: cdict) -> Tuple[cdict, cdict]:
        clean_state = chain_state[-1]
        clean_extra = chain_extra[-1]
        clean_state.alpha = chain_state.alpha.mean()
        clean_extra.parameters = chain_extra.parameters[-1]
        return clean_state, clean_extra

    def forward_proposal_non_zero_weight(self,
                                         abc_scenario: ABCScenario,
                                         state: cdict,
                                         extra: cdict,
                                         random_key: jnp.ndarray) -> cdict:

        def mcmc_kernel(previous_carry: Tuple[cdict, cdict],
                        _: None) -> Tuple[Tuple[cdict, cdict], Tuple[cdict, cdict]]:
            new_carry = self.mcmc_sampler.update(abc_scenario, *previous_carry)
            return new_carry, new_carry

        extra.random_key = random_key

        start_state, start_extra = self.mcmc_sampler.startup(abc_scenario,
                                                             extra.parameters.mcmc_steps,
                                                             state,
                                                             extra)

        final_carry, chain = scan(mcmc_kernel,
                                  (start_state, start_extra),
                                  None,
                                  length=self.parameters.mcmc_steps)

        advanced_state, advanced_extra = self.clean_mcmc_chain(chain[0], chain[1])
        return advanced_state

    def forward_proposal(self,
                         abc_scenario: ABCScenario,
                         state: cdict,
                         extra: cdict,
                         random_key: jnp.ndarray) -> cdict:
        return cond(state.log_weight > -jnp.inf,
                    lambda state_extra: self.forward_proposal_non_zero_weight(abc_scenario, state_extra[0],
                                                                              state_extra[1], random_key),
                    lambda state_extra: state_extra[0],
                    (state, extra))

    def adapt_mcmc_params(self,
                          previous_ensemble_state: cdict,
                          previous_extra: cdict,
                          new_ensemble_state: cdict,
                          new_extra: cdict) -> Tuple[cdict, cdict]:
        return new_ensemble_state, new_extra

    def adapt(self,
              previous_ensemble_state: cdict,
              previous_extra: cdict,
              new_ensemble_state: cdict,
              new_extra: cdict) -> Tuple[cdict, cdict]:
        n = new_ensemble_state.value.shape[0]
        next_threshold = self.next_threshold(new_ensemble_state, new_extra)
        new_ensemble_state.threshold = jnp.ones(n) * next_threshold
        new_extra.parameters.threshold = next_threshold
        new_ensemble_state.log_weight = self.log_weight(previous_ensemble_state, previous_extra,
                                                        new_ensemble_state, new_extra)
        new_ensemble_state.ess = jnp.ones(n) * ess_log_weight(new_ensemble_state.log_weight)
        alive_inds = previous_ensemble_state.log_weight > -jnp.inf
        new_extra.alpha_mean = (new_ensemble_state.alpha * alive_inds).sum() / alive_inds.sum()
        new_ensemble_state, new_extra = self.adapt_mcmc_params(previous_ensemble_state, previous_extra,
                                                               new_ensemble_state, new_extra)

        return new_ensemble_state, new_extra





from mocat.src.abc.scenarios.gk import GKUniformPrior
from mocat.src.abc.scenarios.gk import GKTransformedUniformPrior
from mocat.src.abc.scenarios.gk import GKOnlyAUniformPrior
from mocat.src.abc.scenarios.gk import GKOnlyATransformedUniformPrior

from mocat.src.abc.scenarios.linear_gaussian import LinearGaussian

from mocat.src.abc.scenarios.lotka_volterra import LotkaVolterra
from mocat.src.abc.scenarios.lotka_volterra import TransformedLotkaVolterra
from mocat.src.abc.scenarios.lotka_volterra import lotka_volterra_simulate

from mocat.src.abc.scenarios.SIR import SIR
from mocat.src.abc.scenarios.SIR import TransformedSIR

from mocat.src.abc.scenarios.ricker import Ricker
from mocat.src.abc.scenarios.ricker import ricker_simulate





from mocat.src.abc.scenarios import scenarios


from mocat.src.abc.abc import ABCScenario
from mocat.src.abc.abc import ABCSampler

from mocat.src.abc.importance import ImportanceABC
from mocat.src.abc.importance import VanillaABC

from mocat.src.abc.mcmc import ABCMCMCSampler
from mocat.src.abc.mcmc import RandomWalkABC
from mocat.src.abc.mcmc import RMMetropolisDiagStepsize

from mocat.src.abc.smc import ABCSMCSampler
from mocat.src.abc.smc import MetropolisedABCSMCSampler





########################################################################################################################
# Module: abc/scenarios/gk.py
# Description: Linear Gaussian Bayesian inverse problem.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Union, Any

from jax import numpy as jnp, random

from mocat.src.abc.abc import ABCScenario
from mocat.src.utils import extract_dimension, reset_covariance, gaussian_potential


class LinearGaussian(ABCScenario):
    name = 'Linear Gaussian'

    prior_covariance_sqrt: jnp.ndarray = None
    prior_precision_sqrt: jnp.ndarray = None
    prior_precision_det: float = None
    likelihood_covariance_sqrt: jnp.ndarray = None
    likelihood_precision_sqrt: jnp.ndarray = None
    likelihood_precision_det: float = None

    def __init__(self,
                 prior_mean: Union[float, jnp.ndarray] = 0.,
                 prior_covariance: Union[float, jnp.ndarray] = 1.,
                 data: Union[float, jnp.ndarray] = 0.,
                 likelihood_matrix: Union[float, jnp.ndarray] = 1.,
                 likelihood_covariance: Union[float, jnp.ndarray] = 1.,
                 **kwargs):

        if data is not None:
            self.data = data

        self.dim = extract_dimension(prior_mean, prior_covariance,
                                     likelihood_matrix, likelihood_covariance)
        if self.dim is None:
            raise AttributeError(f'Could not find dimension for {self.__class__.__name__}')

        if isinstance(prior_mean, float):
            prior_mean = jnp.ones(self.dim) * prior_mean
        self.prior_mean = prior_mean

        if isinstance(prior_covariance, float):
            prior_covariance = jnp.eye(self.dim) * prior_covariance
        self.prior_covariance = prior_covariance

        if isinstance(likelihood_matrix, float):
            likelihood_matrix = jnp.eye(self.dim) * likelihood_matrix
        self.dim_obs = likelihood_matrix.shape[0]
        self.likelihood_matrix = likelihood_matrix

        if isinstance(likelihood_covariance, float):
            likelihood_covariance = jnp.eye(self.dim_obs) * likelihood_covariance
        self.likelihood_covariance = likelihood_covariance

        super().__init__(**kwargs)

    def __setattr__(self,
                    key: str,
                    value: Any):
        self.__dict__[key] = value
        if key[-10:] == 'covariance':
            reset_covariance(self, key, value)

    def prior_sample(self,
                     random_key: jnp.ndarray) -> jnp.ndarray:
        return self.prior_mean + self.prior_covariance_sqrt @ random.normal(random_key, (self.dim,))

    def prior_potential(self,
                        x: jnp.ndarray,
                        random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return gaussian_potential(x, self.prior_mean,
                                  sqrt_prec=self.prior_precision_sqrt, det_prec=self.prior_precision_det)

    def likelihood_sample(self,
                          x: jnp.ndarray,
                          random_key: jnp.ndarray) -> jnp.ndarray:
        return self.likelihood_matrix @ x + self.likelihood_covariance_sqrt @ random.normal(random_key, (self.dim_obs,))

    def likelihood_potential(self,
                             x: jnp.ndarray,
                             random_key: jnp.ndarray = None) -> Union[float, jnp.ndarray]:
        return gaussian_potential(self.data, x @ self.likelihood_matrix.T,
                                  sqrt_prec=self.likelihood_precision_sqrt, det_prec=self.likelihood_precision_det)





from mocat.src.ssm.linear_gaussian.linear_gaussian import LinearGaussian
from mocat.src.ssm.linear_gaussian.linear_gaussian import TimeHomogenousLinearGaussian

from mocat.src.ssm.nonlinear_gaussian import NonLinearGaussian

from mocat.src.ssm.scenarios.nonlinear1d import NonLinear1DBenchmark

from mocat.src.ssm.scenarios.stochastic_volatility import StochasticVolatility

from mocat.src.ssm.scenarios.lorenz63 import Lorenz63
from mocat.src.ssm.scenarios.lorenz63 import lorenz63_dynamics
from mocat.src.ssm.scenarios.lorenz63 import lorenz63_integrator

from mocat.src.ssm.scenarios.lorenz96 import Lorenz96
from mocat.src.ssm.scenarios.lorenz96 import lorenz96_dynamics
from mocat.src.ssm.scenarios.lorenz96 import lorenz96_integrator





########################################################################################################################
# Module: twodim/vectorise.py
# Description: Two dimensional scenario class - vectorises potentials (that aren't already)
#              and automatically finds limits for plotting.
#
# Web: https://github.com/SamDuffield/mocat
########################################################################################################################

from typing import Callable, Union, Tuple

import jax.numpy as jnp
from jax import vmap
import matplotlib.pyplot as plt

from mocat.src.core import Scenario


def _find_first_non_zero_row(matrix: jnp.ndim,
                             direction: int = 1):
    if direction == 1:
        i = 0
        while not any(matrix[i]):
            i += 1
        return i
    elif direction == -1:
        i = len(matrix) - 1
        while not any(matrix[i]):
            i += -1
        return i


def _plot_densf(ax, x, y, z, **kwargs):
    return ax.contourf(x, y, z, **kwargs)


class TwoDimToyScenario(Scenario):
    dim = 2
    xlim = None
    ylim = None
    plot_resolution = 1000

    def prior_potential(self,
                        x: jnp.ndarray,
                        random_key: jnp.ndarray = None) -> float:
        # Default to uniform prior unless overwritten (surpresses warning)
        return jnp.zeros_like(jnp.array(x)[..., 0])

    def vec_potential(self,
                      xl: jnp.ndarray,
                      yl: jnp.ndarray,
                      random_key: jnp.ndarray = None):
        return vmap(lambda y: vmap(lambda x: self.potential(jnp.array([x, y]), random_key))(xl))(yl)

    def auto_axes_lims(self,
                       xlim: float = 10,
                       ylim: float = 10.):
        # Assumes self.vec_potential is a vectorised function that goes to 0 in the tails

        # Initial evaluation grid
        ix = jnp.linspace(-xlim, xlim, 100)
        iy = jnp.linspace(-ylim, ylim, 100)

        z = jnp.exp(-self.vec_potential(ix, iy))

        # Find mode
        max_z = jnp.max(z)
        if jnp.isnan(max_z):
            raise TypeError('nan found attempting auto_axes_lims, try giving manual xlim and ylim as kwargs')

        if max_z == 0.:
            return self.auto_axes_lims(xlim=xlim / 1.5, ylim=ylim / 1.5)

        # Area with probability mass
        z_keep = z > max_z / 10

        # Find bounds of area with probability mass
        xlim_new = jnp.array([ix[_find_first_non_zero_row(z_keep.T, direction)] for direction in [1, -1]])
        ylim_new = jnp.array([iy[_find_first_non_zero_row(z_keep, direction)] for direction in [1, -1]])

        if xlim in jnp.abs(xlim_new) or ylim in jnp.abs(ylim_new):
            return self.auto_axes_lims(xlim=2 * xlim if xlim in jnp.abs(xlim_new) else xlim,
                                       ylim=2 * ylim if ylim in jnp.abs(ylim_new) else ylim)

        # Expand
        expansion = 0.05
        xlim += (xlim_new[1] - xlim_new[0]) * expansion * jnp.array([-1, 1])
        ylim += (ylim_new[1] - ylim_new[0]) * expansion * jnp.array([-1, 1])
        self.xlim = tuple(xlim_new)
        self.ylim = tuple(ylim_new)

    def plot(self,
             ax=None,
             xlim=None,
             ylim=None,
             random_key: jnp.ndarray = None,
             recalc_axes=False,
             potential=False,
             cmap='Purples', **kwargs) -> Union[plt.Axes, Tuple[plt.Figure, plt.Axes]]:

        # Check plot bounds
        if ax is not None:
            xlim = ax.get_xlim()
            ylim = ax.get_ylim()
        elif xlim is None or ylim is None:
            # Generate plot dimensions if not already present
            if self.xlim is None or recalc_axes:
                self.auto_axes_lims()
            xlim = self.xlim if xlim is None else xlim
            ylim = self.ylim if ylim is None else ylim

        x_linsp = jnp.linspace(*xlim, self.plot_resolution)
        y_linsp = jnp.linspace(*ylim, self.plot_resolution)
        vals_mat = self.vec_potential(x_linsp, y_linsp, random_key)
        if not potential:
            vals_mat = jnp.exp(-vals_mat)

        if ax is None:
            # Plot contours
            fig, ax = plt.subplots()
            _plot_densf(ax, x_linsp, y_linsp, vals_mat, cmap=cmap, **kwargs)
            return fig, ax
        else:
            return _plot_densf(ax, x_linsp, y_linsp, vals_mat, cmap=cmap, **kwargs)





from mocat.src.scenarios.twodim.vectorise import TwoDimToyScenario

from mocat.src.scenarios.twodim.toy_examples import Gaussian
from mocat.src.scenarios.twodim.toy_examples import Banana
from mocat.src.scenarios.twodim.toy_examples import NealFunnel
from mocat.src.scenarios.twodim.toy_examples import GaussianMixture
from mocat.src.scenarios.twodim.toy_examples import DoubleWell
from mocat.src.scenarios.twodim.toy_examples import Rastrigin

